#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
import datetime

import panedr
try:
    import MDAnalysis
except Exception:
    logging.warn('Required module MDAnalysis not found.')
    MDAnalysis = False

from .metainfo import m_env
from nomad.units import ureg
from nomad.parsing.parser import FairdiParser

from nomad.parsing.file_parser import TextParser, Quantity, FileParser
from nomad.datamodel.metainfo.common_dft import Run, SamplingMethod, System,\
    EnergyContribution, SingleConfigurationCalculation, Topology, Interaction, Method
from .metainfo.gromacs import x_gromacs_section_control_parameters, x_gromacs_section_input_output_files

MOL = 6.022140857e+23


class GromacsLogParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        def str_to_header(val_in):
            val = [v.split(':', 1) for v in val_in.strip().split('\n')]
            return {v[0].strip(): v[1].strip() for v in val if len(v) == 2}

        def str_to_input_parameters(val_in):
            re_array = re.compile(r'\s*([\w\-]+)\[[\d ]+\]\s*=\s*\{*(.+)')
            re_scalar = re.compile(r'\s*([\w\-]+)\s*=\s*(.+)')
            parameters = dict()
            val = val_in.strip().split('\n')
            for val_n in val:
                val_scalar = re_scalar.match(val_n)
                if val_scalar:
                    parameters[val_scalar.group(1)] = val_scalar.group(2)
                    continue
                val_array = re_array.match(val_n)
                if val_array:
                    parameters.setdefault(val_array.group(1), [])
                    value = [float(v) for v in val_array.group(2).rstrip('}').split(',')]
                    parameters[val_array.group(1)].append(value[0] if len(value) == 1 else value)
            return parameters

        def str_to_energies(val_in):
            energy_keys_re = re.compile(r'(.+?)(?:  |\Z| P)')
            keys = []
            values = []
            energies = dict()
            for val in val_in.strip().split('\n'):
                val = val.strip()
                if val[0].isalpha():
                    keys = [k.strip() for k in energy_keys_re.findall(val)]
                    keys = ['P%s' % k if k.startswith('res') else k for k in keys if k]
                else:
                    values = val.split()
                    for n, key in enumerate(keys):

                        if key == 'Temperature':
                            energies[key] = float(values[n]) * ureg.kelvin
                        elif key.startswith('Pres'):
                            key = key.rstrip(' (bar)')
                            energies[key] = float(values[n]) * ureg.bar
                        else:
                            energies[key] = float(values[n]) / MOL * ureg.kJ
            return energies

        def str_to_step_info(val_in):
            val = val_in.strip().split('\n')
            keys = val[0].split()
            values = [float(v) for v in val[1].split()]
            return {key: values[n] for n, key in enumerate(keys)}

        thermo_quantities = [
            Quantity(
                'energies',
                r'Energies \(kJ/mol\)\s*([\s\S]+?)\n\n',
                str_operation=str_to_energies, convert=False),
            Quantity(
                'step_info',
                r'(Step.+\n[\d\.\- ]+)',
                str_operation=str_to_step_info, convert=False)]

        self._quantities = [
            Quantity('time_start', r'Log file opened on (.+)', flatten=False),
            Quantity(
                'host_info',
                r'Host:\s*(\S+)\s*pid:\s*(\d+)\s*rank ID:\s*(\d+)\s*number of ranks:\s*(\d*)'),
            Quantity('module_version', r'GROMACS:\s*(.+?),\s*VERSION\s*(\S+)', flatten=False),
            Quantity('execution_path', r'Executable:\s*(.+)'),
            Quantity('working_path', r'Data prefix:\s*(.+)'),
            # TODO cannot understand treatment of the command line in the old parser
            Quantity(
                'header',
                r'(GROMACS version:[\s\S]+?)\n\n', str_operation=str_to_header),
            Quantity(
                'input_parameters',
                r'Input Parameters:\s*([\s\S]+?)\n\n', str_operation=str_to_input_parameters),
            Quantity(
                'step',
                r'(Step\s*Time[\s\S]+?Energies[\s\S]+?\n\n)',
                repeats=True, sub_parser=TextParser(quantities=thermo_quantities)),
            Quantity(
                'averages',
                r'A V E R A G E S  ====>([\s\S]+?\n\n\n)',
                sub_parser=TextParser(quantities=thermo_quantities)),
            Quantity('time_end', r'Finished \S+ on rank \d+ (.+)', flatten=False)]

    def get_pbc(self):
        pbc = self.get('input_parameters', {}).get('pbc', 'xyz')
        return ['x' in pbc, 'y' in pbc, 'z' in pbc]

    def get_sampling_settings(self):
        input_parameters = self.get('input_parameters', {})
        integrator = input_parameters.get('integrator', 'md').lower()
        if integrator in ['l-bfgs', 'cg', 'steep']:
            sampling_method = 'geometry_optimization'
        elif integrator in ['bd']:
            sampling_method = 'langevin_dynamics'
        else:
            sampling_method = 'molecular_dynamics'

        ensemble_type = 'NVE' if sampling_method == 'molecular_dynamics' else None
        tcoupl = input_parameters.get('tcoupl', 'no').lower()
        if tcoupl != 'no':
            ensemble_type = 'NVT'
            pcoupl = input_parameters.get('pcoupl', 'no').lower()
            if pcoupl != 'no':
                ensemble_type = 'NPT'

        return dict(
            sampling_method=sampling_method, integrator_type=integrator,
            ensemble_type=ensemble_type)

    def get_tpstat_settings(self):
        input_parameters = self.get('input_parameters', {})
        target_T = input_parameters.get('ref-t', 0) * ureg.kelvin

        thermostat_type = None
        tcoupl = input_parameters.get('tcoupl', 'no').lower()
        if tcoupl != 'no':
            thermostat_type = 'Velocity Rescaling' if tcoupl == 'v-rescale' else tcoupl.title()

        thermostat_tau = input_parameters.get('tau-t', 0) * ureg.ps

        # TODO infer langevin_gamma [s] from bd_fric
        # bd_fric = self.get('bd-fric', 0, unit='amu/ps')
        langevin_gamma = None

        target_P = input_parameters.get('ref-p', 0) * ureg.bar
        # if P is array e.g. for non-isotropic pressures, get average since metainfo is float
        if hasattr(target_P, 'shape'):
            target_P = np.average(target_P)

        barostat_type = None
        pcoupl = input_parameters.get('pcoupl', 'no').lower()
        if pcoupl != 'no':
            barostat_type = pcoupl.title()

        barostat_tau = input_parameters.get('tau-p', 0) * ureg.ps

        return dict(
            target_T=target_T, thermostat_type=thermostat_type, thermostat_tau=thermostat_tau,
            target_P=target_P, barostat_type=barostat_type, barostat_tau=barostat_tau,
            langevin_gamma=langevin_gamma)


class GromacsEDRParser(FileParser):
    def __init__(self):
        super().__init__(None)
        self._energy_keys = [
            'LJ (SR)', 'Coulomb (SR)', 'Potential', 'Kinetic En.', 'Total Energy',
            'Vir-XX', 'Vir-XY', 'Vir-XZ', 'Vir-YX', 'Vir-YY', 'Vir-YZ', 'Vir-ZX', 'Vir-ZY',
            'Vir-ZZ']
        self._pressure_keys = [
            'Pressure', 'Pres-XX', 'Pres-XY', 'Pres-XZ', 'Pres-YX', 'Pres-YY', 'Pres-YZ',
            'Pres-ZX', 'Pres-ZY', 'Pres-ZZ']
        self._temperature_keys = ['Temperature']
        self._time_keys = ['Time']

    @property
    def fileedr(self):
        if self._file_handler is None:
            try:
                self._file_handler = panedr.edr_to_df(self.mainfile)
            except Exception:
                self.logger.error('Error reading edr file.')

        return self._file_handler

    def parse(self, key):
        if self.fileedr is None:
            return

        val = self.fileedr.get(key, None)
        if self._results is None:
            self._results = dict()

        if val is not None:
            val = np.asarray(val)
        if key in self._energy_keys:
            val = val / MOL * ureg.kJ
        elif key in self._temperature_keys:
            val = val * ureg.kelvin
        elif key in self._pressure_keys:
            val = val * ureg.bar
        elif key in self._time_keys:
            val = val * ureg.ps

        self._results[key] = val

    def keys(self):
        return list(self.fileedr.keys())

    @property
    def length(self):
        return self.fileedr.shape[0]


class MDAnalysisParser(FileParser):
    def __init__(self):
        super().__init__(None)

    @property
    def trajectory_file(self):
        return self._trajectory_file

    @trajectory_file.setter
    def trajectory_file(self, val):
        self._file_handler = None
        self._trajectory_file = val

    def get_interactions(self):
        interactions = self.get('interactions', None)

        if interactions is not None:
            return interactions

        interaction_types = ['angles', 'bonds', 'dihedrals', 'impropers']
        interactions = []
        for interaction_type in interaction_types:
            try:
                interaction = getattr(self.universe, interaction_type)
            except Exception:
                continue

            for i in range(len(interaction)):
                interactions.append(
                    (str(interaction[i].type), [interaction[i].value()]))

        self._results['interactions'] = interactions

        return interactions

    def get_n_atoms(self, frame_index):
        return self.get('n_atoms', [0] * frame_index)[frame_index]

    def get_cell(self, frame_index):
        return self.get('cell', [np.zeros((3, 3))] * frame_index)[frame_index]

    def get_atom_labels(self, frame_index):
        return self.get('atom_labels', None)

    def get_positions(self, frame_index):
        return self.get('positions', [None] * frame_index)[frame_index]

    def get_velocities(self, frame_index):
        return self.get('velocities', [None] * frame_index)[frame_index]

    def get_forces(self, frame_index):
        return self.get('forces', [None] * frame_index)[frame_index]

    @property
    def universe(self):
        if self._file_handler is None:
            try:
                args = [f for f in [self.trajectory_file] if f is not None]
                self._file_handler = MDAnalysis.Universe(self.mainfile, *args)
            except Exception:
                self.logger.error('Error setting up MDAnalysis.')
        return self._file_handler

    def parse(self, key):
        if self._results is None:
            self._results = dict()

        if self.universe is None:
            return

        if not MDAnalysis:
            return

        atoms = list(self.universe.atoms)
        try:
            trajectory = self.universe.trajectory
        except Exception:
            trajectory = []

        unit = None
        val = None
        if key == 'timestep':
            val = trajectory.dt
            unit = ureg.ps
        elif key == 'atom_labels':
            val = [
                MDAnalysis.topology.guessers.guess_atom_element(atom.name)
                for atom in atoms]
        elif key == 'n_atoms':
            val = [traj.n_atoms for traj in trajectory] if trajectory else [len(atoms)]
        elif key == 'n_frames':
            val = len(trajectory)
        elif key == 'positions':
            val = [traj.positions if traj.has_positions else None for traj in trajectory]
            unit = ureg.angstrom
        elif key == 'velocities':
            val = [traj.velocities if traj.has_velocities else None for traj in trajectory]
            unit = ureg.angstrom / ureg.ps
        elif key == 'forces':
            val = [traj.forces / MOL if traj.has_forces else None for traj in trajectory]
            unit = ureg.kJ / ureg.angstrom
        elif key == 'cell':
            val = [traj.triclinic_dimensions for traj in trajectory]
            unit = ureg.angstrom

        if unit is not None:
            if isinstance(val, list):
                val = [v * unit if v is not None else v for v in val]
            else:
                val = val * unit

        self._results[key] = val


class GromacsParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/gromacs', code_name='Gromacs', code_homepage='http://www.gromacs.org/',
            domain='dft', mainfile_contents_re=r'gmx mdrun, (VERSION|version)')
        self._metainfo_env = m_env
        self.log_parser = GromacsLogParser()
        self.traj_parser = MDAnalysisParser()
        self.energy_parser = GromacsEDRParser()
        self._metainfo_mapping = {
            'LJ (SR)': 'Leonard-Jones', 'Coulomb (SR)': 'coulomb',
            'Potential': 'potential', 'Kinetic En.': 'kinetic'}

    def get_gromacs_file(self, ext):
        files = [d for d in self._gromacs_files if d.endswith(ext)]

        if len(files) == 1:
            return os.path.join(self._maindir, files[0])

        # we assume that the file has the same basename as the log file e.g.
        # out.log would correspond to out.tpr and out.trr and out.edr
        for f in files:
            if f.rsplit('.', 1)[0] == self._basename:
                return os.path.join(self._maindir, f)

        for f in files:
            if f.rsplit('.', 1)[0].startswith(self._basename):
                return os.path.join(self._maindir, f)

        # if the files are all named differently, we guess that the one that does not
        # share the same basename would be file we are interested in
        # e.g. in a list of files out.log someout.log out.tpr out.trr another.tpr file.trr
        # we guess that the out.* files belong together and the rest that does not share
        # a basename would be grouped together
        counts = []
        for f in files:
            count = 0
            for reff in self._gromacs_files:
                if f.rsplit('.', 1)[0] == reff.rsplit('.', 1)[0]:
                    count += 1
            if count == 1:
                return os.path.join(self._maindir, f)
            counts.append(count)

        return os.path.join(self._maindir, files[counts.index(min(counts))])

    def parse_thermodynamic_data(self):
        sec_run = self.archive.section_run[-1]

        forces = self.traj_parser.get('forces')
        for n, forces_n in enumerate(forces):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)
            sec_scc.atom_forces = forces_n
            sec_scc.single_configuration_calculation_to_system_ref = sec_run.section_system[n]
            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]

        # get it from edr file
        if len(self.energy_parser.keys()) > 0:
            thermo_data = self.energy_parser
            n_evaluations = self.energy_parser.length
        else:
            # try to get it from log file
            steps = self.log_parser.get('step', [])
            thermo_data = dict()
            for n, step in enumerate(steps):
                n = int(step.get('step_info', {}).get('Step', n))
                if step.energies is None:
                    continue
                keys = step.energies.keys()
                for key in keys:
                    thermo_data.setdefault(key, [None] * len(forces))
                    thermo_data[key][n] = step.energies.get(key)
                info = step.get('step_info', {})
                thermo_data.setdefault('Time', [None] * len(forces))
                thermo_data['Time'][n] = info.get('Time', None)
            n_evaluations = n + 1

        create_scc = False
        if len(forces) != n_evaluations:
            self.logger.warn(
                'Mismatch in number of calculations and number of thermodynamic'
                'evaluations, will create new sections')
            create_scc = True

        timestep = self.traj_parser.get('timestep')
        if timestep is None:
            timestep = self.log_parser.get('input_parameters', {}).get('dt', 1.0) * ureg.ps

        # TODO add other energy contributions, properties
        energy_keys = ['LJ (SR)', 'Coulomb (SR)', 'Potential', 'Kinetic En.']

        for n in range(n_evaluations):
            if create_scc:
                sec_scc = sec_run.m_create(SingleConfigurationCalculation)
            else:
                sec_scc = sec_run.section_single_configuration_calculation[n]

            for key in thermo_data.keys():
                val = thermo_data.get(key)[n]
                if val is None:
                    continue

                if key == 'Total Energy':
                    sec_scc.energy_total = val

                elif key == 'Pressure':
                    sec_scc.pressure = val

                elif key == 'Temperature':
                    sec_scc.temperature = val

                elif key == 'Time':
                    sec_scc.time_step = int((val / timestep).magnitude)
                if key in energy_keys:
                    sec_energy = sec_scc.m_create(EnergyContribution)
                    sec_energy.energy_contribution_kind = self._metainfo_mapping[key]
                    sec_energy.energy_contribution_value = val

    def parse_topology(self):
        sec_run = self.archive.section_run[-1]

        sec_topology = sec_run.m_create(Topology)
        try:
            n_atoms = self.traj_parser.get('n_atoms', [0])[0]
        except Exception:
            gro_file = self.get_gromacs_file('gro')
            self.traj_parser.mainfile = gro_file
            n_atoms = self.traj_parser.get('n_atoms', [0])[0]

        sec_topology.number_of_topology_atoms = n_atoms

        interactions = self.traj_parser.get_interactions()
        for interaction in interactions:
            if not interaction[0] or not interaction[1]:
                continue
            sec_interaction = sec_topology.m_create(Interaction)
            sec_interaction.interaction_kind = interaction[0]
            sec_interaction.interaction_parameters = interaction[1]

        sec_run.section_method[-1].method_to_topology_ref = sec_topology

    def parse_system(self):
        sec_run = self.archive.section_run[-1]

        n_frames = self.traj_parser.get('n_frames', 0)

        pbc = self.log_parser.get_pbc()
        for n in range(n_frames):
            positions = self.traj_parser.get_positions(n)
            sec_system = sec_run.m_create(System)
            if positions is None:
                continue

            sec_system.number_of_atoms = self.traj_parser.get_n_atoms(n)
            sec_system.configuration_periodic_dimensions = pbc
            sec_system.simulation_cell = self.traj_parser.get_cell(n)
            sec_system.lattice_vectors = self.traj_parser.get_cell(n)
            sec_system.atom_labels = self.traj_parser.get_atom_labels(n)
            sec_system.atom_positions = positions

            velocities = self.traj_parser.get_velocities(n)
            if velocities is not None:
                sec_system.atom_velocities = velocities

    def parse_method(self):
        self.archive.section_run[-1].m_create(Method)
        # TODO what to put here

    def parse_sampling_method(self):
        sec_run = self.archive.section_run[-1]
        sec_sampling_method = sec_run.m_create(SamplingMethod)

        sampling_settings = self.log_parser.get_sampling_settings()

        sec_sampling_method.sampling_method = sampling_settings['sampling_method']
        sec_sampling_method.ensemble_type = sampling_settings['ensemble_type']
        sec_sampling_method.x_gromacs_integrator_type = sampling_settings['integrator_type']

        input_parameters = self.log_parser.get('input_parameters', {})
        timestep = input_parameters.get('dt', 0)
        sec_sampling_method.x_gromacs_integrator_dt = timestep

        nsteps = input_parameters.get('nsteps', 0)
        sec_sampling_method.x_gromacs_number_of_steps_requested = nsteps

        tp_settings = self.log_parser.get_tpstat_settings()

        target_T = tp_settings.get('target_T', None)
        if target_T is not None:
            sec_sampling_method.x_gromacs_thermostat_target_temperature = target_T
        thermostat_tau = tp_settings.get('thermostat_tau', None)
        if thermostat_tau is not None:
            sec_sampling_method.x_gromacs_thermostat_tau = thermostat_tau
        target_P = tp_settings.get('target_P', None)
        if target_P is not None:
            sec_sampling_method.x_gromacs_barostat_target_pressure = target_P
        barostat_tau = tp_settings.get('barostat_P', None)
        if barostat_tau is not None:
            sec_sampling_method.x_gromacs_barostat_tau = barostat_tau
        langevin_gamma = tp_settings.get('langevin_gamma', None)
        if langevin_gamma is not None:
            sec_sampling_method.x_gromacs_langevin_gamma = langevin_gamma

    def parse_input(self):
        sec_run = self.archive.section_run[-1]
        sec_input_output_files = sec_run.m_create(x_gromacs_section_input_output_files)

        topology_file = os.path.basename(self.traj_parser.mainfile)
        if topology_file.endswith('tpr'):
            sec_input_output_files.x_gromacs_inout_file_topoltpr = topology_file
        elif topology_file.endswith('gro'):
            sec_input_output_files.x_gromacs_inout_file_confoutgro = topology_file

        trajectory_file = os.path.basename(self.traj_parser.trajectory_file)
        sec_input_output_files.x_gromacs_inout_file_trajtrr = trajectory_file

        edr_file = os.path.basename(self.energy_parser.mainfile)
        sec_input_output_files.x_gromacs_inout_file_eneredr = edr_file

        sec_control_parameters = sec_run.m_create(x_gromacs_section_control_parameters)
        input_parameters = self.log_parser.get('input_parameters', {})
        input_parameters.update(self.log_parser.get('header', {}))
        for key, val in input_parameters.items():
            key = 'x_gromacs_inout_control_%s' % key.replace('-', '').replace(' ', '_').lower()
            if hasattr(sec_control_parameters, key):
                val = str(val) if not isinstance(val, np.ndarray) else val
                setattr(sec_control_parameters, key, val)

    def init_parser(self):
        self.log_parser.mainfile = self.filepath
        self.log_parser.logger = self.logger
        self.traj_parser.logger = self.logger
        self.energy_parser.logger = self.logger

    def reuse_parser(self, parser):
        self.log_parser.quantities = parser.log_parser.quantities

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self._maindir = os.path.dirname(self.filepath)
        self._gromacs_files = os.listdir(self._maindir)
        self._basename = os.path.basename(filepath).rsplit('.', 1)[0]

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program_name = 'GROMACS'

        header = self.log_parser.get('header', {})
        sec_run.program_version = header.get('GROMACS version', 'unknown').lstrip('VERSION ')

        for key in ['start', 'end']:
            time = self.log_parser.get('time_%s')
            if time is None:
                continue
            setattr(sec_run, 'time_run_date_%s' % key, datetime.datetime.strptime(
                time, '%a %b %d %H:%M:%S %Y').timestamp())

        host_info = self.log_parser.get('host_info')
        if host_info is not None:
            sec_run.x_gromacs_program_execution_host = host_info[0]
            sec_run.x_gromacs_parallel_task_nr = host_info[1]
            sec_run.x_gromacs_number_of_tasks = host_info[2]

        self.parse_method()

        self.parse_sampling_method()

        topology_file = self.get_gromacs_file('tpr')
        # I have no idea if output trajectory file can be specified in input
        trajectory_file = self.get_gromacs_file('trr')
        self.traj_parser.mainfile = topology_file
        self.traj_parser.trajectory_file = trajectory_file

        self.parse_topology()

        self.parse_system()

        # TODO read also from ene
        edr_file = self.get_gromacs_file('edr')
        self.energy_parser.mainfile = edr_file

        self.parse_thermodynamic_data()

        self.parse_input()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask import Blueprint, Response, request
from flask_restplus import Api, reqparse
import urllib.parse
from rdflib import Graph

from nomad import config

blueprint = Blueprint('dcat', __name__)

base_url = config.api_url(api='dcat')


def url(*args, **kwargs):
    ''' Returns the full dcat api url for the given path (args) and query (kwargs) parameters. '''
    url = f'{base_url.rstrip("/")}/{"/".join(args).lstrip("/")}'

    if len(kwargs) > 0:
        return f'{url}?{urllib.parse.urlencode(kwargs)}'
    else:
        return url


api = Api(
    blueprint,
    version='1.0', title='NOMAD\'s API for servicing dcat resources',
    description='NOMAD\'s API for serving dcat resources',
    validate=True)


# For some unknown reason it is necessary for each fr api to have a handler.
# Otherwise the global app error handler won't be called.
@api.errorhandler(Exception)
def errorhandler(error):
    '''When an internal server error is caused by an unexpected exception.'''
    return str(error)


arg_parser = reqparse.RequestParser()
arg_parser.add_argument('format', type=str, choices=[
    'xml',
    'n3',
    'turtle',
    'nt',
    'pretty-xml',
    'trig'])

all_repsonse_types = {
    'application/xml': 'xml',
    'application/rdf+prettyxml': 'pretty-xml',
    'application/rdf': 'xml',
    'application/rdf+xml': 'xml',
    'text/plain': 'n3',
    'text/turtle': 'turtle',
    'text/nt': 'nt',
    'text/n3': 'n3',
    'text/rdf+n3': 'n3',
    'text/rdf+nt': 'nt',
    'text/rdf+turtle': 'turtle',
    'application/x-trig': 'trig'
}

response_types = [
    'application/xml',
    'application/rdf+xml',
    'application/rdf+pretty-xml',
    'text/plain',
    'text/turtle',
    'text/rdf+n3',
    'text/rdf+nt',
    'application/x-trig']


def rdf_respose(g: Graph) -> Response:
    args = arg_parser.parse_args()
    format_ = args.get('format')
    if format_ is None:
        accept_header = request.headers.get('Accept', None)
        if accept_header is not None:
            format_ = all_repsonse_types.get(accept_header, 'pretty-xml')
        else:
            format_ = 'pretty-xml'

    try:
        content_type = next(key for key, value in all_repsonse_types.items() if value == format_)
    except StopIteration:
        content_type = 'application/xml' if format in ['xml', 'pretty-xml'] else 'text/%s' % format_

    return Response(
        g.serialize(format=format_).decode('utf-8'), 200,
        {'Content-Type': content_type})



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.metainfo.legacy import LegacyDefinition

from vaspparser.metainfo import vasp_incars
from nomad.datamodel.metainfo import public

m_package = Package(
    name='vasp_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='vasp.nomadmetainfo.json'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    vasp_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        build date as string
        ''',
        categories=[public.program_info, public.accessory_info],
        a_legacy=LegacyDefinition(name='vasp_build_date'))

    vasp_src_date = Quantity(
        type=str,
        shape=[],
        description='''
        date of last modification of the source as string
        ''',
        categories=[public.program_info, public.accessory_info],
        a_legacy=LegacyDefinition(name='vasp_src_date'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_vasp_incar_in = Quantity(
        type=JSON,
        shape=[],
        description='''
        contains all the user-input INCAR parameters
        ''',
        a_legacy=LegacyDefinition(name='x_vasp_incar_in'))

    x_vasp_incar_out = Quantity(
        type=JSON,
        shape=[],
        description='''
        contains the actual INCAR parameters used by VASP at runtime
        ''',
        a_legacy=LegacyDefinition(name='x_vasp_incar_out'))

    x_vasp_unknown_incars = Quantity(
        type=JSON,
        shape=[],
        description='''
        INCAR variables uknown wrt to Vasp Wiki
        ''',
        a_legacy=LegacyDefinition(name='x_vasp_unknown_incars'))

    x_vasp_atom_kind_refs = Quantity(
        type=public.section_method_atom_kind,
        shape=['number_of_atoms'],
        description='''
        reference to the atom kinds of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_vasp_atom_kind_refs'))

    x_vasp_k_points_generation_method = Quantity(
        type=str,
        shape=[],
        description='''
        k points generation  method
        ''',
        categories=[public.settings_potential_energy_surface, vasp_incars.x_vasp_incar_param, public.settings_k_points],
        a_legacy=LegacyDefinition(name='x_vasp_k_points_generation_method'))

    x_vasp_k_points = Quantity(
        type=int,
        shape=['x_vasp_number_of_k_points'],
        description='''
        k points
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_k_points'))

    x_vasp_numer_of_magmom = Quantity(
        type=int,
        shape=[],
        description='''
        number of magnetic moments, number_of_atoms for ISPIN = 2, 3*number of atoms for
        non-collinear magnetic systems
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_numer_of_magmom'))

    x_vasp_tetrahedrons_list = Quantity(
        type=np.dtype(np.int32),
        shape=['N', 5],
        description='''
        Rows of 5 elements. First the weight (symmetry degeneration), then the four corner
        points of each tetrahedron.
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_tetrahedrons_list'))

    x_vasp_tetrahedron_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[1],
        description='''
        Volume weight of a single tetrahedron (all tetra's must have the same volume)
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_tetrahedron_volume'))

    x_vasp_nose_thermostat = Quantity(
        type=np.dtype(np.float64),
        shape=[4],
        description='''
        Nose thermostat output
        ''',
        categories=[vasp_incars.x_vasp_incar_param],
        a_legacy=LegacyDefinition(name='x_vasp_nose_thermostat'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_vasp_selective_dynamics = Quantity(
        type=np.dtype(bool),
        shape=['number_of_atoms', 3],
        description='''
        Boolean array to eiter allow or forbid coordinate modifications during relaxation
        ''',
        a_legacy=LegacyDefinition(name='x_vasp_selective_dynamics'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import numpy as np
from ase.io.trajectory import Trajectory

from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import FileParser
from nomad.datamodel.metainfo.common_dft import Run, Topology, Constraint, Method, System,\
    SingleConfigurationCalculation, SamplingMethod


class TrajParser(FileParser):
    def __init__(self):
        super().__init__()

    @property
    def traj(self):
        if self._file_handler is None:
            self._file_handler = Trajectory(self.mainfile, 'r')
            # check if traj file is really asap
            if hasattr(self._file_handler.backend, 'calculator'):
                if self._file_handler.backend.calculator.name != 'emt':
                    self.logger.error('Trajectory is not ASAP.')
                    self._file_handler = None
        return self._file_handler

    def get_version(self):
        if hasattr(self.traj, 'ase_version') and self.traj.ase_version:
            return self.traj.ase_version
        else:
            return '3.x.x'


class AsapParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/asap', code_name='ASAP', domain='dft',
            mainfile_name_re=r'.*.traj$', mainfile_mime_re=r'application/octet-stream')
        self.traj_parser = TrajParser()

    def init_parser(self):
        self.traj_parser.mainfile = self.filepath
        self.traj_parser.logger = self.logger

    def parse_topology(self):

        def get_constraint_name(constraint):
            def index():
                d = constraint['kwargs'].get('direction')
                return ((d / np.linalg.norm(d)) ** 2).argsort()[2]

            name = constraint.get('name')
            if name == 'FixedPlane':
                return ['fix_yz', 'fix_xz', 'fix_xy'][index()]
            elif name == 'FixedLine':
                return ['fix_x', 'fix_y', 'fix_z'][index()]
            elif name == 'FixAtoms':
                return 'fix_xyz'
            else:
                return name

        traj = self.traj_parser.traj[0]
        sec_topology = self.archive.section_run[0].m_create(Topology)
        sec_topology.topology_force_field_name = traj.calc.name
        for constraint in traj.constraints:
            sec_constraint = sec_topology.m_create(Constraint)
            as_dict = constraint.todict()
            indices = as_dict['kwargs'].get('a', as_dict['kwargs'].get('indices'))
            sec_constraint.constraint_atoms = np.asarray(indices)
            sec_constraint.constraint_kind = get_constraint_name(as_dict)

    def parse_system(self, traj):
        sec_system = self.archive.section_run[0].m_create(System)

        sec_system.lattice_vectors = traj.get_cell() * ureg.angstrom
        sec_system.atom_labels = traj.get_chemical_symbols()
        sec_system.atom_positions = traj.get_positions() * ureg.angstrom
        sec_system.configuration_periodic_dimensions = traj.get_pbc()
        if traj.get_velocities() is not None:
            sec_system.atom_velocities = traj.get_velocities() * (ureg.angstrom / ureg.fs)

    def parse_scc(self, traj):
        sec_scc = self.archive.section_run[0].m_create(SingleConfigurationCalculation)

        try:
            sec_scc.energy_total = traj.get_total_energy() * ureg.eV
        except Exception:
            pass

        try:
            sec_scc.atom_forces = traj.get_forces() * (ureg.eV / ureg.angstrom)
        except Exception:
            pass

        try:
            sec_scc.atom_forces_raw = traj.get_forces(apply_constraint=False) * (ureg.eV / ureg.angstrom)
        except Exception:
            pass

    def parse_method(self):
        traj = self.traj_parser.traj

        sec_method = self.archive.section_run[0].m_create(Method)
        sec_method.calculation_method = traj[0].calc.name

        description = traj.description if hasattr(traj, 'description') else dict()

        if not description:
            return

        sec_sampling_method = self.archive.section_run[0].m_create(SamplingMethod)

        for key in ['timestep', 'maxstep']:
            val = description.get(key)
            if val is not None:
                setattr(sec_sampling_method, 'x_asap_%s' % key, val)

        calc_type = description.get('type')
        if calc_type == 'optimization':
            sec_sampling_method.sampling_method = 'geometry_optimization'
            sec_sampling_method.geometry_optimization_method = description.get('optimizer', '').lower()
        elif calc_type == 'molecular-dynamics':
            sec_sampling_method.sampling_method = 'molecular_dynamics'
            sec_sampling_method.x_asap_temperature = description.get('temperature', 0)

        md_type = description.get('md-type', '')
        if 'Langevin' in md_type:
            sec_sampling_method.ensemble_type = 'NVT'
            sec_sampling_method.x_asap_langevin_friction = description.get('friction', 0)
        elif 'NVT' in md_type:
            sec_sampling_method.ensemble_type = 'NVT'
        elif 'Verlet' in md_type:
            sec_sampling_method.ensemble_type = 'NVE'
        elif 'NPT' in md_type:
            sec_sampling_method.ensemble_type = 'NPT'

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        if self.traj_parser.traj is None:
            return

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'ASAP'
        sec_run.program_version = self.traj_parser.get_version()

        # TODO do we build the topology and method for each frame
        self.parse_topology()
        self.parse_method()
        for traj in self.traj_parser.traj:
            self.parse_system(traj)
            self.parse_scc(traj)
            # add references to scc
            sec_scc = sec_run.section_single_configuration_calculation[-1]
            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]
            sec_scc.single_configuration_calculation_to_system_ref = sec_run.section_system[-1]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Iterator, Any
from types import FunctionType
import urllib
import sys
import inspect
from fastapi import Request, Query, HTTPException  # pylint: disable=unused-import
from pydantic import ValidationError, BaseModel  # pylint: disable=unused-import
import zipstream

if sys.version_info >= (3, 7):
    import zipfile
else:
    import zipfile37 as zipfile  # pragma: no cover


def parameter_dependency_from_model(name: str, model_cls):
    '''
    Takes a pydantic model class as input and creates a dependency with corresponding
    Query parameter definitions that can be used for GET
    requests.

    This will only work, if the fields defined in the input model can be turned into
    suitable query parameters. Otherwise fastapi will complain down the road.

    Arguments:
        name: Name for the dependency function.
        model_cls: A ``BaseModel`` inheriting model class as input.
    '''
    names = []
    annotations: Dict[str, type] = {}
    defaults = []
    for field_model in model_cls.__fields__.values():
        field_info = field_model.field_info

        names.append(field_model.name)
        annotations[field_model.name] = field_model.outer_type_
        defaults.append(Query(field_model.default, description=field_info.description))

    code = inspect.cleandoc('''
    def %s(%s):
        try:
            return %s(%s)
        except ValidationError as e:
            errors = e.errors()
            for error in errors:
                error['loc'] = ['query'] + list(error['loc'])
            raise HTTPException(422, detail=errors)

    ''' % (
        name, ', '.join(names), model_cls.__name__,
        ', '.join(['%s=%s' % (name, name) for name in names])))

    compiled = compile(code, 'string', 'exec')
    env = {model_cls.__name__: model_cls}
    env.update(**globals())
    func = FunctionType(compiled.co_consts[0], env, name)
    func.__annotations__ = annotations
    func.__defaults__ = (*defaults,)

    return func


class File(BaseModel):
    path: str
    f: Any
    size: int


def create_streamed_zipfile(
        files: Iterator[File],
        compress: bool = False) -> Iterator[bytes]:

    '''
    Creates a streaming zipfile object that can be used in fastapi's ``StreamingResponse``.
    '''

    def path_to_write_generator():
        for file_obj in files:
            def content_generator():
                while True:
                    data = file_obj.f.read(1024 * 64)
                    if not data:
                        break
                    yield data

            yield dict(
                arcname=file_obj.path,
                iterable=content_generator(),
                buffer_size=file_obj.size)

    compression = zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
    zip_stream = zipstream.ZipFile(mode='w', compression=compression, allowZip64=True)
    zip_stream.paths_to_write = path_to_write_generator()

    for chunk in zip_stream:
        yield chunk


def create_responses(*args):
    return {
        status_code: response
        for status_code, response in args}


def update_url_query_arguments(original_url: str, **kwargs) -> str:
    '''
    Takes an url, and returns a new url, obtained by updating the query arguments in the
    `original_url` as specified by the kwargs.
    '''
    scheme, netloc, path, params, query, fragment = urllib.parse.urlparse(original_url)
    query_dict = urllib.parse.parse_qs(query)
    for k, v in kwargs.items():
        if v is None:
            # Unset the value
            if k in query_dict:
                query_dict.pop(k)
        else:
            query_dict[k] = [str(v)]
    query = urllib.parse.urlencode(query_dict, doseq=True)
    return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Callable, Any, Dict, List, DefaultDict
from collections import defaultdict

from nomad import config, metainfo
from nomad.metainfo.elastic_extension import Elastic


search_quantities_by_index: DefaultDict[str, Dict[str, 'Search']] = defaultdict(dict)
''' All available search quantities by their full qualified name. '''

search_sub_sections_by_index: DefaultDict[str, Dict[str, 'Search']] = defaultdict(dict)
''' All available sub sections in the search index with full qualified name. '''

metrics_by_index: DefaultDict[str, Dict[str, 'Search']] = defaultdict(dict)
'''
The available search metrics. Metrics are integer values given for each entry that can
be used in statistics (aggregations), e.g. the sum of all total energy calculations or
cardinality of all unique geometries. First key is the index name, second key
is the metric name.
'''

groups_by_index: DefaultDict[str, Dict[str, 'Search']] = defaultdict(dict)
''' The available groupable quantities. First key is the index name, second key
is the metric name. '''

order_default_quantities_by_index: DefaultDict[str, Dict[str, 'Search']] = defaultdict(dict)
''' The quantity for each domain (key) that is the default quantity to order search results by. '''


search_quantities = search_quantities_by_index[config.elastic.index_name]
search_sub_sections = search_sub_sections_by_index[config.elastic.index_name]
groups = groups_by_index[config.elastic.index_name]
metrics = metrics_by_index[config.elastic.index_name]
order_default_quantities = order_default_quantities_by_index[config.elastic.index_name]


# TODO multi, split are more flask related
class Search(Elastic):
    '''
    A metainfo quantity annotation class that defines additional properties that determine
    how to search for the respective quantity. Only quantities that have this will
    be mapped to elastic search. The annotation is an extension of :class:`Elastic` and
    add nomad API specific search features like grouping, statistics, metrics, domains, etc.

    Attributes:
        name: The name of this search quantity. Will be the name in the elastic index and
            the name for the search parameter. Default is the metainfo quantity name.
        many_or: Indicates that an 'or' (es terms) search is performed if many values are given.
            Otherwise an 'and' (es bool->should->match) is performed.  Values are 'split' and
            'append' to indicate how URL search parameters should be treated.
        many_and: Indicates that many values can be supplied for search. Values are 'split' and
            'append' to indicate how URL search parameters should be treated.
        order_default: Indicates that this quantity is used to order search results
            if no other ordering was specificed.
        metric: Quantity can be used to build statistics. Statistics provide a metric
            value for each value of the quantity. E.g. number of datasets with a given atom label.
            This defines a metric based on this quantity. Values need to be a valid
            elastic search aggregation (e.g. sum, cardinality, etc.).
        metric_name: If this quantity is indicated to function as a metric, the metric
            needs a name. By default the quantities name is used.
        statistics_size:
            The maximum number of values in a statistic. Default is 10.
        statistics_order:
            The order key that is passed to elastic search to determine the order of
            the statistic values.
        statistics_values:
            If the statistics has a fixed set of values, use this parameter to set it
            as a list of strings. Will fill statistics_size with the len of this list.
            The information can be used (e.g. by the GUI) to fill in empty values.
        group: Indicates that his quantity can be used to group results. The value will
            be the name of the group.
        derived: A callable that is applied to search parameter values before search.
        search_field: The qualified field in the elastic mapping that is used to search.
            This might be different from the field that is used to store the value in
            elastic search. This is especially useful if the field represents a inner
            document and a subfield of this inner object should be used for search.
        nested: Indicates if a subsection should be treated as a Nested field.
            Defaults to False meaning that the subsection is treated as an inner
            object.
    '''

    def __init__(
            self,
            name: str = None, description: str = None,
            many_and: str = None, many_or: str = None,
            order_default: bool = False,
            group: str = None, metric: str = None, metric_name: str = None,
            statistic_size: int = None,
            statistic_order: str = '_key',
            statistic_values: List[str] = None,
            derived: Callable[[Any], Any] = None,
            search_field: str = None,
            nested: bool = False,
            **kwargs):

        super().__init__(field=None, **kwargs)

        self.name = name
        self.description = description
        self.many_and = many_and
        self.many_or = many_or
        self.order_default = order_default
        self.group = group
        self.metric = metric
        self.metric_name = metric_name

        self.statistic_fixed_size = statistic_size
        self.statistic_size = statistic_size if statistic_size is not None else 20
        self.statistic_order = statistic_order
        self.statistic_values = statistic_values
        self.search_field = search_field
        self.nested = nested

        self.derived = derived

        self.qualified_name: str = None

        assert many_and is None or many_or is None, 'A search quantity can only be used for multi or many search'
        assert many_and in [None, 'split', 'append'], 'Only split and append are valid values'
        assert many_or in [None, 'split', 'append'], 'Only split and append are valid values'

    def init_annotation(self, definition):
        if self.name is None:
            self.name = definition.name
        assert self.name is not None

        if self.description is None:
            self.description = definition.description

        super().init_annotation(definition)

    def register(self, prefix, field, index):
        domain_or_all = self.definition.m_parent.m_get_annotations('domain', '__all__')

        prefix_and_dot = prefix + '.' if prefix is not None else ''

        self.qualified_name = prefix_and_dot + self.name
        if self.search_field is not None:
            self.search_field = prefix_and_dot + self.search_field
        else:
            self.search_field = self.qualified_name

        if self.definition.m_def == metainfo.SubSection.m_def:
            if not self.nested:
                search_sub_sections_by_index[index][self.qualified_name] = self
                return

        assert self.qualified_name not in search_quantities_by_index[index], 'Search quantities must have a unique name: %s' % self.name
        search_quantities_by_index[index][self.qualified_name] = self

        if self.metric is not None:
            if self.metric_name is None:
                self.metric_name = self.qualified_name
            else:
                self.metric_name = prefix_and_dot + self.metric_name

            assert self.metric_name not in metrics_by_index[index], 'Metric names must be unique: %s' % self.metric_name
            metrics_by_index[index][self.metric_name] = self

        if self.group is not None:
            self.group = prefix_and_dot + self.group
            assert self.group not in groups_by_index[index], 'Groups must be unique'
            groups_by_index[index][self.group] = self

        if self.order_default:
            assert order_default_quantities_by_index[index].get(domain_or_all) is None, 'Only one quantity can be the order default'
            order_default_quantities_by_index[index][domain_or_all] = self

    @property
    def argparse_action(self):
        if self.many_or is not None:
            return self.many_or

        if self.many_and is not None:
            return self.many_and

        return None

    @property
    def many(self) -> bool:
        return self.many_and is not None or self.many_or is not None

    @property
    def flask_field(self):
        from flask_restplus import fields
        value_field = fields.String
        if self.definition.type == int:
            value_field = fields.Integer

        if self.many:
            return fields.List(value_field(), description=self.description)
        else:
            return value_field(description=self.description)

    @property
    def statistic_values(self):
        return self._statistic_values

    @statistic_values.setter
    def statistic_values(self, values):
        self._statistic_values = values
        if self._statistic_values is not None:
            if self.statistic_fixed_size is None:
                self.statistic_size = len(self._statistic_values)
            self.statistic_fixed_size = len(self._statistic_values)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module describes all configurable parameters for the nomad python code. The
configuration is used for all executed python code including API, worker, CLI, and other
scripts. To use the configuration in your own scripts or new modules, simply import
this module.

All parameters are structured into objects for two reasons. First, to have
categories. Second, to allow runtime manipulation that is not effected
by python import logic. The categories are choosen along infrastructure components:
``mongo``, ``elastic``, etc.

This module also provides utilities to read the configuration from environment variables
and .yaml files. This is done automatically on import. The precedence is env over .yaml
over defaults.

.. autoclass:: nomad.config.NomadConfig
.. autofunction:: nomad.config.load_config
'''

import logging
import os
import os.path
import yaml
import warnings

try:
    from nomad import gitinfo
except ImportError:
    git_root = os.path.join(os.path.dirname(__file__), '..')
    cwd = os.getcwd()
    os.chdir(git_root)
    os.system('./gitinfo.sh')
    os.chdir(cwd)

    from nomad import gitinfo


warnings.filterwarnings('ignore', message='numpy.dtype size changed')
warnings.filterwarnings('ignore', message='numpy.ufunc size changed')


class NomadConfig(dict):
    '''
    A class for configuration categories. It is a dict subclass that uses attributes as
    key/value pairs.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


CELERY_WORKER_ROUTING = 'worker'
CELERY_QUEUE_ROUTING = 'queue'

rabbitmq = NomadConfig(
    host='localhost',
    user='rabbitmq',
    password='rabbitmq'
)


def rabbitmq_url():
    return 'pyamqp://%s:%s@%s//' % (rabbitmq.user, rabbitmq.password, rabbitmq.host)


celery = NomadConfig(
    max_memory=64e6,  # 64 GB
    timeout=1800,  # 1/2 h
    acks_late=False,
    routing=CELERY_QUEUE_ROUTING,
    priorities={
        'Upload.process_upload': 5,
        'Upload.delete_upload': 9,
        'Upload.publish_upload': 10
    }
)

fs = NomadConfig(
    tmp='.volumes/fs/tmp',
    staging='.volumes/fs/staging',
    public='.volumes/fs/public',
    local_tmp='/tmp',
    prefix_size=2,
    archive_version_suffix=None,
    working_directory=os.getcwd()
)

elastic = NomadConfig(
    host='localhost',
    port=9200,
    index_name='nomad_fairdi_calcs',
    materials_index_name='nomad_fairdi_materials'
)

keycloak = NomadConfig(
    server_url='https://nomad-lab.eu/fairdi/keycloak/auth/',
    realm_name='fairdi_nomad_prod',
    username='admin',
    password='*',
    client_id='nomad_public',
    client_secret=None,
    oasis=False)

mongo = NomadConfig(
    host='localhost',
    port=27017,
    db_name='nomad_fairdi'
)

logstash = NomadConfig(
    enabled=False,
    host='localhost',
    tcp_port='5000',
    level=logging.DEBUG
)

services = NomadConfig(
    api_host='localhost',
    api_port=8000,
    api_base_path='/fairdi/nomad/latest',
    api_secret='defaultApiSecret',
    api_chaos=0,
    admin_user_id='00000000-0000-0000-0000-000000000000',
    not_processed_value='not processed',
    unavailable_value='unavailable',
    https=False,
    https_upload=False,
    upload_limit=10,
    force_raw_file_decoding=False,
    download_scan_size=500,
    download_scan_timeout=u'30m'
)

oasis = NomadConfig(
    central_nomad_api_url='https://nomad-lab.eu/prod/rae/api',
    central_nomad_deployment_id='nomad-lab.eu/prod/rae',
    allowed_users=None  # a list of usernames or user account emails
)

tests = NomadConfig(
    default_timeout=30
)


def api_url(ssl: bool = True, api: str = 'api'):
    protocol = 'https' if services.https and ssl else 'http'
    host_and_port = services.api_host.strip('/')
    if services.api_port not in [80, 443]:
        host_and_port += ':' + str(services.api_port)
    base_path = services.api_base_path.strip('/')
    return f'{protocol}://{host_and_port}/{base_path}/{api}'


def gui_url(page: str = None):
    base = api_url(True)[:-3]
    if base.endswith('/'):
        base = base[:-1]

    if page is not None:
        return '%s/gui/%s' % (base, page)

    return '%s/gui' % base


def _check_config():
    """Used to check that the current configuration is valid. Should only be
    called once after the final config is loaded.

    Raises:
        AssertionError: if there is a contradiction or invalid values in the
            config file settings.
    """
    # The AFLOW symmetry information is checked once on import
    proto_symmetry_tolerance = normalize.prototype_symmetry_tolerance
    symmetry_tolerance = normalize.symmetry_tolerance
    if proto_symmetry_tolerance != symmetry_tolerance:
        raise AssertionError(
            "The AFLOW prototype information is outdated due to changed tolerance "
            "for symmetry detection. Please update the AFLOW prototype information "
            "by running the CLI command 'nomad admin ops prototype-update "
            "--matches-only'."
        )

    if normalize.springer_db_path and not os.path.exists(normalize.springer_db_path):
        normalize.springer_db_path = None


mail = NomadConfig(
    enabled=False,
    with_login=False,
    host='',
    port=8995,
    user='',
    password='',
    from_address='support@nomad-lab.eu',
    cc_address='support@nomad-lab.eu'
)

normalize = NomadConfig(
    # The system size limit for running the dimensionality analysis. For very
    # large systems the dimensionality analysis will get too expensive.
    system_classification_with_clusters_threshold=64,
    # Symmetry tolerance controls the precision used by spglib in order to find
    # symmetries. The atoms are allowed to move 1/2*symmetry_tolerance from
    # their symmetry positions in order for spglib to still detect symmetries.
    # The unit is angstroms. The value of 0.1 is used e.g. by Materials Project
    # according to
    # https://pymatgen.org/pymatgen.symmetry.analyzer.html#pymatgen.symmetry.analyzer.SpacegroupAnalyzer
    symmetry_tolerance=0.1,
    # The symmetry tolerance used in aflow prototype matching. Should only be
    # changed before re-running the prototype detection.
    prototype_symmetry_tolerance=0.1,
    # Maximum number of atoms in the single cell of a 2D material for it to be
    # considered valid.
    max_2d_single_cell_size=7,
    # The distance tolerance between atoms for grouping them into the same
    # cluster. Used in detecting system type.
    cluster_threshold=2.5,
    # Defines the "bin size" for rounding cell angles for the material hash
    angle_rounding=float(10.0),  # unit: degree
    # The threshold for a system to be considered "flat". Used e.g. when
    # determining if a 2D structure is purely 2-dimensional to allow extra rigid
    # transformations that are improper in 3D but proper in 2D.
    flat_dim_threshold=0.1,
    # The threshold for point equality in k-space. Unit: 1/m.
    k_space_precision=150e6,
    # The energy threshold for how much a band can be on top or below the fermi
    # level in order to detect a gap. Unit: Joule.
    band_structure_energy_tolerance=1.6022e-20,  # 0.1 eV
    springer_db_path=os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'normalizing/data/springer.msg'
    )
)

paths = NomadConfig(
    similarity="",
)

client = NomadConfig(
    user='leonard.hofstadter@nomad-fairdi.tests.de',
    password='password',
    url='http://nomad-lab.eu/prod/rae/api'
)

datacite = NomadConfig(
    mds_host='https://mds.datacite.org',
    enabled=False,
    prefix='10.17172',
    user='*',
    password='*'
)

meta = NomadConfig(
    version='0.10.11',
    commit=gitinfo.commit,
    release='devel',
    deployment='standard',
    default_domain='dft',
    service='unknown nomad service',
    name='novel materials discovery (NOMAD)',
    description='A FAIR data sharing platform for materials science data',
    homepage='https://nomad-lab.eu',
    source_url='https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-FAIR',
    maintainer_email='markus.scheidgen@physik.hu-berlin.de',
    deployment_id='nomad-lab.eu/prod/rae',
    beta=None
)

gitlab = NomadConfig(
    private_token='not set'
)

auxfile_cutoff = 100
parser_matching_size = 150 * 80  # 150 lines of 80 ASCII characters per line
console_log_level = logging.WARNING
max_upload_size = 32 * (1024 ** 3)
raw_file_strip_cutoff = 1000
max_entry_download = 500000
use_empty_parsers = False
reprocess_match = False
reprocess_unmatched = True
reprocess_rematch = True
process_reuse_parser = True
metadata_file_name = 'nomad'
metadata_file_extensions = ('json', 'yaml', 'yml')
enable_lazy_import = True
encyclopedia_enabled = True
aitoolkit_enabled = False


def normalize_loglevel(value, default_level=logging.INFO):
    plain_value = value
    if plain_value is None:
        return default_level
    else:
        try:
            return int(plain_value)
        except ValueError:
            return getattr(logging, plain_value)


_transformations = {
    'console_log_level': normalize_loglevel,
    'logstash_level': normalize_loglevel
}


# use std python logger, since logging is not configured while loading configuration
logger = logging.getLogger(__name__)


def _apply(key, value, raise_error: bool = True) -> None:
    '''
    Changes the config according to given key and value. The first part of a key
    (with ``_`` as a separator) is interpreted as a group of settings. E.g. ``fs_staging``
    leading to ``config.fs.staging``.
    '''
    full_key = key
    try:
        group_key, config_key = full_key.split('_', 1)
    except Exception:
        if raise_error:
            logger.error(f'config key does not exist: {full_key}')
        return

    current = globals()

    if group_key not in current:
        if key not in current:
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return
    else:
        current = current[group_key]
        if not isinstance(current, NomadConfig):
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return

        if config_key not in current:
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return

        key = config_key

    try:
        current_value = current[key]
        if current_value is not None and not isinstance(value, type(current_value)):
            value = _transformations.get(full_key, type(current_value))(value)

        current[key] = value
        logger.info(f'set config setting {full_key}={value}')
    except Exception as e:
        logger.error(f'cannot set config setting {full_key}={value}: {e}')


def _apply_env_variables():
    kwargs = {
        key[len('NOMAD_'):].lower(): value
        for key, value in os.environ.items()
        if key.startswith('NOMAD_') and key != 'NOMAD_CONFIG'}

    for key, value in kwargs.items():
        _apply(key, value, raise_error=False)


def _apply_nomad_yaml():
    config_file = os.environ.get('NOMAD_CONFIG', 'nomad.yaml')

    if not os.path.exists(config_file):
        return

    with open(config_file, 'r') as stream:
        try:
            config_data = yaml.load(stream, Loader=getattr(yaml, 'FullLoader'))
        except yaml.YAMLError as e:
            logger.error(f'cannot read nomad config: {e}')
            return

    for key, value in config_data.items():
        if isinstance(value, dict):
            group_key = key
            for key, value in value.items():
                _apply(f'{group_key}_{key}', value)
        else:
            _apply(key, value)


def load_config():
    '''
    Loads the configuration from nomad.yaml and environment.
    '''
    _apply_nomad_yaml()
    _apply_env_variables()
    _check_config()


load_config()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os.path
import json
import ase
import re
import numpy as np
from datetime import datetime

from nomad.parsing.parser import FairdiParser
from nomad.datamodel.metainfo.common_experimental import (
    Experiment, Data, Method, Sample, Material, Location)


class MPESParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/mpes', code_name='mpes', code_homepage='https://github.com/mpes-kit/mpes',
            domain='ems', mainfile_mime_re=r'(application/json)|(text/.*)', mainfile_name_re=(r'.*.meta'),
            mainfile_contents_re=(r'"data_repository_name": "zenodo.org"')
        )

    def parse(self, filepath, archive, logger=None):
        with open(filepath, 'rt') as f:
            data = json.load(f)

        experiment = archive.m_create(Experiment)
        experiment.raw_metadata = data

        # Read general experimental parameters
        # experiment.experiment_location = ', '.join(reversed(re.findall(r"[\w']+", data.get('experiment_location'))))
        location = experiment.m_create(Location)
        location.address = data.get('experiment_location')
        location.institution = data.get('facility_institution')
        location.facility = data.get('facility')

        dates = data.get('experiment_date')
        if dates:
            start, end = dates.split(' ')
            try:
                experiment.experiment_time = datetime.strptime(start, '%m.%Y')
            except ValueError:
                pass
            try:
                experiment.experiment_end_time = datetime.strptime(end, '%m.%Y')
            except ValueError:
                pass
        experiment.experiment_summary = data.get('experiment_summary')

        # Read data parameters
        section_data = experiment.m_create(Data)
        section_data.repository_name = data.get('data_repository_name')
        section_data.entry_repository_url = data.get('data_repository_url')
        section_data.repository_url = '/'.join(data.get('data_repository_url').split('/')[0:3])
        section_data.preview_url = 'preview.png'

        # Read method parameters
        method = experiment.m_create(Method)
        method.data_type = 'multidimensional spectrum'
        method.method_name = data.get('experiment_method')
        method.method_abbreviation = data.get('experiment_method_abbrv')
        method.instrument_description = data.get('equipment_description')
        method.probing_method = 'laser pulses'
        method.general_beamline = data.get('beamline')
        method.general_source_pump = data.get('source_pump')
        method.general_source_probe = data.get('source_probe')
        method.general_measurement_axis = np.array(re.findall(r"[\w']+", data.get('measurement_axis')))
        method.general_physical_axis = np.array(re.findall(r"[\w']+", data.get('physical_axis')))

        # Read parameters related to experimental source
        # source_gid = backend.openSection('experiment_source_parameters')
        method.source_pump_repetition_rate = data.get('pump_rep_rate')
        method.source_pump_pulse_duration = data.get('pump_pulse_duration')
        method.source_pump_wavelength = data.get('pump_wavelength')
        method.source_pump_spectrum = np.array(data.get('pump_spectrum'))
        method.source_pump_photon_energy = data.get('pump_photon_energy')
        method.source_pump_size = np.array(data.get('pump_size'))
        method.source_pump_fluence = np.array(data.get('pump_fluence'))
        method.source_pump_polarization = data.get('pump_polarization')
        method.source_pump_bunch = data.get('pump_bunch')
        method.source_probe_repetition_rate = data.get('probe_rep_rate')
        method.source_probe_pulse_duration = data.get('probe_pulse_duration')
        method.source_probe_wavelength = data.get('probe_wavelength')
        method.source_probe_spectrum = np.array(data.get('probe_spectrum'))
        method.source_probe_photon_energy = data.get('probe_photon_energy')
        method.source_probe_size = np.array(data.get('probe_size'))
        method.source_probe_fluence = np.array(data.get('probe_fluence'))
        method.source_probe_polarization = data.get('probe_polarization')
        method.source_probe_bunch = data.get('probe_bunch')
        method.source_temporal_resolution = data.get('temporal_resolution')

        # Read parameters related to detector
        # detector_gid = backend.openSection('experiment_detector_parameters')
        method.detector_extractor_voltage = data.get('extractor_voltage')
        method.detector_work_distance = data.get('work_distance')
        method.detector_lens_names = np.array(re.findall(r"[\w']+", data.get('lens_names')))
        method.detector_lens_voltages = np.array(data.get('lens_voltages'))
        method.detector_tof_distance = data.get('tof_distance')
        method.detector_tof_voltages = np.array(data.get('tof_voltages'))
        method.detector_sample_bias = data.get('sample_bias')
        method.detector_magnification = data.get('magnification')
        method.detector_voltages = np.array(data.get('detector_voltages'))
        method.detector_type = data.get('detector_type')
        method.detector_sensor_size = np.array(data.get('sensor_size'))
        method.detector_sensor_count = data.get('sensor_count')
        method.detector_sensor_pixel_size = np.array(data.get('sensor_pixel_size'))
        method.detector_calibration_x_to_momentum = np.array(data.get('calibration_x_to_momentum'))
        method.detector_calibration_y_to_momentum = np.array(data.get('calibration_y_to_momentum'))
        method.detector_calibration_tof_to_energy = np.array(data.get('calibration_tof_to_energy'))
        method.detector_calibration_stage_to_delay = np.array(data.get('calibration_stage_to_delay'))
        method.detector_calibration_other_converts = np.array(data.get('calibration_other_converts'))
        method.detector_momentum_resolution = np.array(data.get('momentum_resolution'))
        method.detector_spatial_resolution = np.array(data.get('spatial_resolution'))
        method.detector_energy_resolution = np.array(data.get('energy_resolution'))

        # Read parameters related to sample
        sample = experiment.m_create(Sample)
        sample.sample_description = data.get('sample_description')
        sample.sample_id = data.get('sample_id')
        sample.sample_state_of_matter = data.get('sample_state')
        sample.sample_purity = data.get('sample_purity')
        sample.sample_surface_termination = data.get('sample_surface_termination')
        sample.sample_layers = data.get('sample_layers')
        sample.sample_stacking_order = data.get('sample_stacking_order')
        sample.sample_chemical_id_cas = data.get('chemical_id_cas')
        sample.sample_temperature = data.get('sample_temperature')
        sample.sample_pressure = data.get('sample_pressure')
        sample.sample_growth_method = data.get('growth_method')
        sample.sample_preparation_method = data.get('preparation_method')
        sample.sample_vendor = data.get('sample_vendor')
        sample.sample_substrate_material = data.get('substrate_material')
        sample.sample_substrate_state_of_matter = data.get('substrate_state')
        sample.sample_substrate_vendor = data.get('substrate_vendor')

        material = sample.m_create(Material)
        material.space_group = data.get('sample_space_group')
        material.chemical_name = data.get('chemical_name')
        material.chemical_formula = data.get('chemical_formula')
        atoms = set(ase.Atoms(data.get('chemical_formula')).get_chemical_symbols())
        material.atom_labels = np.array(list(atoms))

        # TODO sample classification
        sample.sample_microstructure = 'bulk sample, polycrystalline'
        sample.sample_constituents = 'multi phase'



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class MolcasParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        def get_positions(val):
            try:
                labels, positions = zip(*re.findall(rf'\d+ +([A-Z][a-z]*)\d* +({re_f} +{re_f} +{re_f})', val))
                positions = np.array([v.split() for v in positions], dtype=np.dtype(np.float64)) * ureg.angstrom
            except Exception:
                labels, positions = [], []
            return dict(atom_labels=labels, atom_positions=positions)

        super().__init__(
            specifications=dict(
                name='parsers/molcas', code_name='MOLCAS', code_homepage='http://www.molcas.org/',
                domain='dft', mainfile_contents_re=r'M O L C A S'),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            # include code name to distinguish gamess and firefly
            program_version=r'version ([\d\.]+ patchlevel \d+)',
            atom_labels_atom_positions=(r'No\. +Label +X +Y +Z\s*([\s\S]+?)\n *\n', get_positions),
            energy_total=rf'Total.* energy *\:*\=* *({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from flask_restplus import Resource, abort
from elasticsearch.exceptions import NotFoundError

from nomad import search
from nomad.app.flask.api.auth import authenticate

from .api import api, arg_parser, rdf_respose, response_types
from .mapping import Mapping

ns = api.namespace('datasets', description='The API for DCAT datasets.')


@ns.route('/<string:entry_id>')
class Dataset(Resource):
    @api.doc('get_dcat_dataset')
    @api.expect(arg_parser)
    @api.representation('application/xml')
    @api.produces(response_types)
    @api.response(404, 'There is no entry with the given id.')
    @api.response(401, 'This entry is not publically accessible.')
    @api.response(200, 'Data send', headers={'Content-Type': 'application/xml'})
    @authenticate()
    def get(self, entry_id):
        ''' Returns a DCAT dataset for a given NOMAD entry id. '''
        try:
            entry = search.entry_document.get(entry_id)
        except NotFoundError:
            abort(404, message='There is no calculation with id %s' % entry_id)

        if entry.with_embargo or not entry.published:
            abort(401, message='Not authorized to access %s' % entry_id)

        mapping = Mapping()
        mapping.map_entry(entry)
        return rdf_respose(mapping.g)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import io
import requests
import click
import typing
import sys
import bravado.exception
import json

from nomad import config, utils
from nomad import files
from nomad.cli import parse as cli_parse

from .client import client


class CalcProcReproduction:
    '''
    Instances represent a local reproduction of the processing for a single calculation.
    It allows to download raw data from a nomad server and reproduce its processing
    (parsing, normalizing) with the locally installed parsers and normalizers.

    The use-case is error/warning reproduction. Use ELK to identify errors, use
    the upload, archive ids to given by ELK, and reproduce and fix the error
    in your development environment.

    Arguments:
        calc_id: The calc_id of the calculation to locally process.
        override: Set to true to override any existing local calculation data.
    '''
    def __init__(self, archive_id: str, override: bool = False, mainfile: str = None) -> None:
        if '/' in archive_id:
            self.calc_id = utils.archive.calc_id(archive_id)
            self.upload_id = utils.archive.upload_id(archive_id)
        else:
            self.calc_id = archive_id
            self.upload_id = 'unknown'

        self.logger = utils.get_logger(__name__, upload_id=self.upload_id, calc_id=self.calc_id)

        self.mainfile = mainfile
        self.parser = None

        from nomad.cli.client import create_client
        client = create_client()
        if self.mainfile is None:
            try:
                calc = client.repo.get_repo_calc(
                    upload_id=self.upload_id, calc_id=self.calc_id).response().result
            except bravado.exception.HTTPNotFound:
                self.logger.error(
                    'could not find calculation, maybe it was deleted or the ids are wrong')
                sys.exit(1)
            self.mainfile = calc['mainfile']
            self.upload_id = calc['upload_id']
            self.logger = self.logger.bind(upload_id=self.upload_id, mainfile=self.mainfile)
        else:
            self.logger = self.logger.bind(mainfile=self.mainfile)
            self.logger.info('Using provided mainfile', mainfile=self.mainfile)

        local_path = os.path.join(config.fs.tmp, 'repro_%s.zip' % archive_id)
        if not os.path.exists(os.path.dirname(local_path)):
            os.makedirs(os.path.dirname(local_path))

        if not os.path.exists(local_path) or override:
            # download raw if not already downloaded or if override is set
            # download with request, since bravado does not support streaming
            self.logger.info('Downloading calc.', mainfile=self.mainfile)
            try:
                token = client.auth.get_auth().response().result.signature_token
                dir_name = os.path.dirname(self.mainfile)
                req = requests.get(
                    '%s/raw/%s/%s/*?signature_token=%s' % (config.client.url, self.upload_id, dir_name, token),
                    stream=True, headers={})
                with open(local_path, 'wb') as f:
                    for chunk in req.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE):
                        f.write(chunk)
            except bravado.exception.HTTPNotFound:
                self.logger.error(
                    'could not find calculation, maybe it was deleted or the ids are wrong')
                sys.exit(1)
            except Exception as e:
                self.logger.error('could not download calculation', exc_info=e)
                sys.exit(1)
        else:
            self.logger.info('Calc already downloaded.')

        self.upload_files = files.ArchiveBasedStagingUploadFiles(
            upload_id='tmp_%s' % archive_id, upload_path=local_path, create=True,
            is_authorized=lambda: True)

    def __enter__(self):
        # open/extract upload file
        self.logger.info('Extracting calc data.')
        try:
            self.upload_files.extract()
        except AssertionError:
            # already extracted
            pass

        assert self.mainfile is not None
        self.logger = self.logger.bind(mainfile=self.mainfile)
        self.logger.info('Identified mainfile.')

        return self

    def __exit__(self, *args):
        self.upload_files.delete()

    def parse(self, parser_name: str = None, **kwargs):
        '''
        Run the given parser on the downloaded calculation. If no parser is given,
        do parser matching and use the respective parser.
        '''
        return cli_parse.parse(
            self.upload_files.raw_file_object(self.mainfile).os_path,
            parser_name=parser_name, logger=self.logger, **kwargs)

    def normalize(self, normalizer: typing.Union[str, typing.Callable], entry_archive=None):
        '''
        Parse the downloaded calculation and run the given normalizer.
        '''
        if entry_archive is None:
            entry_archive = self.parse()

        return cli_parse.normalize(entry_archive=entry_archive, normalizer=normalizer, logger=self.logger)

    def normalize_all(self, entry_archive=None):
        '''
        Parse the downloaded calculation and run the whole normalizer chain.
        '''
        return cli_parse.normalize_all(entry_archive=entry_archive, logger=self.logger)


@client.command(help='Run processing locally.')
@click.argument('CALC_ID', nargs=1, required=True, type=str)
@click.option('--override', is_flag=True, help='Override existing local calculation data.')
@click.option('--show-archive', is_flag=True, help='Print the archive data.')
@click.option('--show-metadata', is_flag=True, help='Print the extracted repo metadata.')
@click.option('--mainfile', default=None, type=str, help='Use this mainfile (in case mainfile cannot be retrived via API.')
@click.option('--skip-normalizers', is_flag=True, help='Do not normalize.')
@click.option('--not-strict', is_flag=True, help='Also match artificial parsers.')
def local(calc_id, show_archive, show_metadata, skip_normalizers, not_strict, **kwargs):
    utils.get_logger(__name__).info('Using %s' % config.client.url)

    with CalcProcReproduction(calc_id, **kwargs) as local:
        if local.upload_id != 'unknown':
            print(
                'Data being saved to .volumes/fs/tmp/repro_'
                '%s if not already there' % local.upload_id)
        entry_archive = local.parse(strict=not not_strict)

        if not skip_normalizers:
            local.normalize_all(entry_archive=entry_archive)

        if show_archive:
            json.dump(entry_archive.m_to_dict(), sys.stdout, indent=2)

        if show_metadata:
            metadata = entry_archive.section_metadata
            metadata.apply_domain_metadata(entry_archive)
            json.dump(metadata.m_to_dict(), sys.stdout, indent=4)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import numpy as np
from xml.etree import ElementTree
from lxml import etree

from nomad.parsing.file_parser import FileParser


class XMLParser(FileParser):
    '''
    Parser for XML files using ElementTree.

    Arguments:
        mainfile: the file to be parsed
        logger: logger
        convert: specifies if quantities are converted automatically
    '''
    def __init__(self, mainfile: str = None, logger=None, **kwargs):
        super().__init__(mainfile, logger=logger, open=kwargs.get('open', None))
        self.convert = kwargs.get('convert', True)
        self.init_parameters()

    def init_parameters(self):
        '''
        Method to call after loading the xml file.
        '''
        self._elements = None

    @property
    def root(self):
        '''
        Returns the root of the XML tree.
        '''
        if self._file_handler is None:
            if self.mainfile is None:
                return
            try:
                self._file_handler = ElementTree.parse(self.mainfile_obj).getroot()
            except Exception:
                try:
                    # I cannot use the lxml XMLParser directly because it is not compatible with
                    # the ElementTree implementation.
                    xml = etree.parse(self.mainfile_mainfile_obj, parser=etree.XMLParser(recover=True))
                    self._file_handler = ElementTree.fromstring(etree.tostring(xml))
                except Exception:
                    pass

                self.logger.error('failed to load xml file')
                try:
                    # I cannot use the lxml XMLParser directly because it is not compatible with
                    # the ElementTree implementation.
                    xml = etree.parse(self.open(self.mainfile), parser=etree.XMLParser(recover=True))
                    self._file_handler = ElementTree.fromstring(etree.tostring(xml))
                except Exception:
                    pass
            self.init_parameters()

        return self._file_handler

    @property
    def elements(self):
        '''
        Returns a list of all elements in the XML tree.
        '''
        if self._elements is None:
            self._elements = self.root.findall('.//')

        return self._elements

    def parse(self, key, convert=None):
        '''
        Parse a quantity identified by key or an xpath-style path. Automatic conversion
        can be switch off by setting convert to False.
        '''
        _convert = convert if convert is not None else self._kwargs.get('convert', None)
        _convert = _convert if _convert is not None else self.convert
        if self._results is None:
            self._results = dict()

        if not self.root:
            return self

        key_in = key
        key = key.lstrip('/')
        if key.find('/') > 0:
            parent = os.path.dirname(key)
            child = os.path.basename(key)
            elements = self.root.findall(os.path.join('./', parent))
        else:
            elements = self.elements
            child = key

        val = []
        for element in elements:
            if child:
                v = element.attrib.get(child, None)
                if v is None:
                    v = element.findall(child)
                    v = [e.text if e.text is not None else '' for e in v]
                if v:
                    val.append(v)

            else:
                val.append(element.attrib)

        if not val:
            return self

        def convert_value(val_in):
            if isinstance(val_in, dict):
                val = dict()
                for k, v in val_in.items():
                    val[k] = convert_value(v)

            elif isinstance(val_in, str):
                # exponential formatting
                val = val_in.strip().split()
                if len(val) == 1:
                    val = val[0]
                    re_float = r'(\d*\.*\d*)d(\-*\+*\d+)'
                    val = re.sub(re_float, r'\1e\2', val)
                    if val.isdecimal():
                        val = int(val)
                    elif val == 'true' or val == 'false':
                        val = val == 'true'
                    else:
                        try:
                            val = float(val)
                        except Exception:
                            pass
                else:
                    val = [convert_value(v) for v in val]

            elif isinstance(val_in, list):
                try:
                    val = [v.split() if isinstance(v, str) else v for v in val_in]
                    val = [v[0] if (isinstance(v, list) and len(v) == 1) else v for v in val]
                    val = np.array(val, dtype=float)
                    if np.all(np.mod(val, 1) == 0):
                        val = np.array(val, dtype=int)
                except Exception:
                    val = [convert_value(v) for v in val_in]

            return val

        if _convert:
            val = convert_value(val)

        val = val[0] if len(val) == 1 else val

        self._results[key_in] = val
        return self



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fastapi import Depends, APIRouter, status

from nomad.utils import strip

from .auth import create_user_dependency
from ..models import User, HTTPExceptionModel
from ..utils import create_responses

router = APIRouter()
default_tag = 'users'


_authentication_required_response = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. The operation requires authorization,
        but no or bad authentication credentials are given.''')}


@router.get(
    '/me',
    tags=[default_tag],
    summary='Get your account data',
    description='Returnes the account data of the authenticated user.',
    responses=create_responses(_authentication_required_response),
    response_model=User)
async def read_users_me(current_user: User = Depends(create_user_dependency(required=True))):
    return current_user



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import numpy as np

from .metainfo import m_env
from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity

from nomad.datamodel.metainfo.common_dft import Run, Method, SingleConfigurationCalculation,\
    BasisSet, System, XCFunctionals, MethodToMethodRefs, ScfIteration, Eigenvalues, Dos,\
    SamplingMethod, ExcitedStates


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\-\+\d\.Ee]+'

        self._energy_mapping = {
            'Total Energy': 'energy_total', 'Nuclear Repulsion': 'nuc_repulsion',
            'Electronic Energy': 'elec_energy', 'One Electron Energy': 'one_elec_energy',
            'Two Electron Energy': 'two_elec_energy', 'Potential Energy': 'potential_energy',
            'Kinetic Energy': 'kinetc_energy', r'E\(X\)': 'exchange_energy',
            r'E\(C\)': 'correlation_energy', r'E\(XC\)': 'exchange_correlation_energy'}

        self._timing_mapping = {
            'Total time': 'final_time', 'Sum of individual times': 'sum_individual_times',
            'Fock matrix formation': 'fock_matrix_formation', 'Coulomb formation': 'coulomb_formation',
            r'Split\-RI-J': 'split_rj', 'XC integration': 'xc_integration',
            r'Basis function eval\.': 'basis_fn_evaluation', r'Density eval\.': 'density_evaluation',
            r'XC\-Functional eval\.': 'xc_functional_evaluation', r'XC\-Potential eval\.': 'potential_evaluation',
            'Diagonalization': 'diagonalization', 'Density matrix formation': 'density_matrix_formation',
            'Population analysis': 'population_analysis', 'Initial guess': 'initial_guess',
            'Orbital Transformation': 'orbital_transformation', 'Orbital Orthonormalization': 'orbital_orthonormalization',
            'DIIS solution': 'diis_solution', 'Grid generation': 'grid_generation'}

        def str_to_cartesian_coordinates(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            symbols = [v[0][:2] for v in val]
            coordinates = np.array([v[1:4] for v in val], dtype=float)
            return symbols, coordinates * ureg.angstrom

        basis_set_quantities = [
            Quantity('basis_set_atom_labels', r'Type\s*(\w+)', repeats=True),
            Quantity('basis_set', r':\s*(\w+)\s*contracted\s*to', repeats=True),
            Quantity('basis_set_contracted', r'(\w+)\s*pattern', repeats=True)]

        basis_set_statistics_quantities = [
            Quantity(
                'nb_of_primitive_gaussian_shells',
                r'# of primitive gaussian shells\s*\.+\s*(\d+)', repeats=True, dtype=int),
            Quantity(
                'nb_of_primitive_gaussian_functions',
                r'# of primitive gaussian functions\s*\.+\s*(\d+)', repeats=True, dtype=int),
            Quantity(
                'nb_of_contracted_shells',
                r'# of contracted shells\s*\.+\s*(\d+)', repeats=True, dtype=int),
            Quantity(
                'nb_of_contracted_basis_functions',
                r'# of contracted (?:aux-)?basis functions\s*\.+\s*(\d+)', repeats=True, dtype=int),
            Quantity(
                'highest_angular_moment',
                r'Highest angular momentum\s*\.+\s*(\d+)', repeats=True, dtype=int),
            Quantity(
                'maximum_contraction_depth',
                r'Maximum contraction depth\s*\.+\s*(\d+)', repeats=True, dtype=int)]

        grid_quantities = [
            Quantity(
                'gral_integ_accuracy',
                rf'General Integration Accuracy\s*IntAcc\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'radial_grid_type',
                r'Radial Grid Type\s*RadialGrid\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'angular_grid',
                r'Angular Grid \(max\. acc\.\)\s*AngularGrid\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'grid_pruning_method',
                r'Angular grid pruning method\s*GridPruning\s*\.+\s*(.+)',
                flatten=False, convert=False),
            Quantity(
                'weight_gener_scheme',
                r'Weight generation scheme\s*WeightScheme\s*\.+\s*(\w+)', convert=False),
            Quantity(
                'basis_fn_cutoff',
                rf'Basis function cutoff\s*BFCut\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'integr_weight_cutoff',
                rf'Integration weight cutoff\s*WCut\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'nb_grid_pts_after_initial_pruning',
                r'# of grid points \(after initial pruning\)\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_grid_pts_after_weights_screening',
                r'# of grid points \(after weights\+screening\)\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'total_nb_grid_pts',
                r'Total number of grid points\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'total_nb_batches',
                r'Total number of batches\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'avg_nb_points_per_batch',
                r'Average number of points per batch\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'avg_nb_grid_pts_per_atom',
                r'Average number of grid points per atom\s*\.+\s*(\d+)', dtype=int)]

        scf_convergence_quantities = [
            Quantity(
                name.lower().replace(' ', '_').replace('-', '_'),
                rf'%s\s*\.+\s*({re_float})\s* Tolerance :\s*({re_float})' % name,
                dtype=float, unit=ureg.hartree) for name in [
                    'Last Energy change', 'Last MAX-Density change', 'Last RMS-Density change']]

        population_quantities = [
            Quantity(
                'atomic_charges',
                r'[A-Z]+ ATOMIC CHARGES.*\n\-+([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=[
                    Quantity('species', r'\n *\d+\s*(\w+)', repeats=True),
                    Quantity('charge', rf':\s*({re_float})', repeats=True, dtype=float),
                    Quantity(
                        'total_charge',
                        rf'Sum of atomic charges\s*:\s*({re_float})', dtype=float)])),
            Quantity(
                'orbital_charges',
                r'[A-Z]+ REDUCED ORBITAL CHARGES.*\s*\-+([\s\S]+?\n\n)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'atom',
                        r'([A-Z][a-z]?\s*[spdf][\s\S]+?)\n *(?:\d|\Z)',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('species', r'([A-Z][a-z]?)', convert=False),
                            Quantity('charge', rf'([spdf]\S*)\s*:\s*({re_float})', repeats=True)]))]))]

        self_consistent_quantities = [
            Quantity(
                'scf_settings',
                r'SCF SETTINGS\s*\-+([\s\S]+?)\-{10}', sub_parser=TextParser(quantities=[
                    Quantity(
                        'XC_functional_type',
                        r'Ab initio Hamiltonian\s*Method\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'XC_functional_type',
                        r'Density Functional\s*Method\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'exchange_functional',
                        r'Exchange Functional\s*Exchange\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'xalpha_param',
                        rf'X-Alpha parameter\s*XAlpha\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'beckes_beta_param',
                        rf'Becke\'s b parameter\s*XBeta\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'correl_functional',
                        r'Correlation Functional Correlation\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'lda_part_of_gga_corr',
                        r'LDA part of GGA corr\.\s*LDAOpt\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'scalar_relativistic_method',
                        r'Scalar relativistic method\s*\.+\s*(\w+)', convert=False),
                    Quantity(
                        'speed_of_light_used',
                        rf'Speed of light used\s*Velit\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'hf_type',
                        r'Hartree-Fock type\s*HFTyp\s*\.+\s*(\w+)', convert=False),
                    Quantity(
                        'total_charge',
                        rf'Total Charge\s*Charge\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'multiplicity',
                        rf'Multiplicity\s*Mult\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'nelectrons',
                        rf'Number of Electrons\s*NEL\s*\.+\s*({re_float})', dtype=float),
                    Quantity(
                        'nuclear_repulsion',
                        rf'Nuclear Repulsion\s*ENuc\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
                    Quantity(
                        'convergence_check_mode',
                        r'Convergence Check Mode ConvCheckMode\s*\.+\s*(\S+)', convert=False),
                    Quantity(
                        'energy_change_tolerance',
                        rf'Energy Change\s*TolE\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
                    Quantity(
                        '1_elect_energy_change',
                        rf'1\-El\. energy change\s*\.+\s*({re_float})', dtype=float)])),
            Quantity(
                'dft_grid_generation',
                r'DFT GRID GENERATION\s*\-+([\s\S]+?\-{10})',
                sub_parser=TextParser(quantities=grid_quantities)),
            Quantity(
                'scf_iterations',
                r'SCF ITERATIONS\s*\-+([\s\S]+?)\*{10}',
                sub_parser=TextParser(quantities=[Quantity(
                    'energy',
                    rf'\n *\d+\s*({re_float})', repeats=True, dtype=float, unit=ureg.hartree)])),
            Quantity(
                'final_grid',
                r'Setting up the final grid:([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=grid_quantities)),
            Quantity(
                'total_scf_energy',
                r'TOTAL SCF ENERGY\s*\-+([\s\S]+?)\-{10}', sub_parser=TextParser(quantities=[
                    Quantity(
                        name,
                        rf'%s\s*:\s*({re_float})' % key, dtype=float, unit=ureg.hartree)
                    for key, name in self._energy_mapping.items()] + [
                        Quantity(
                            'virial_ratio',
                            rf'Virial Ratio\s*:\s*({re_float})', dtype=float),
                        Quantity(
                            'nb_elect_alpha_channel',
                            rf'N\(Alpha\)\s*:\s*({re_float})', dtype=float),
                        Quantity(
                            'nb_elect_beta_channel',
                            rf'N\(Beta\)\s*:\s*({re_float})', dtype=float),
                        Quantity(
                            'nb_elect_total',
                            rf'N\(Total\)\s*:\s*({re_float})', dtype=float)])),
            Quantity(
                'scf_convergence',
                r'SCF CONVERGENCE\s*\-+([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=scf_convergence_quantities)),
            Quantity(
                'orbital_energies',
                r'NO\s*OCC\s*E\(Eh\)\s*E\(eV\)\s*([\s\S]+?)\n\n',
                str_operation=lambda x: np.array([v.split()[:4] for v in x.split('\n')], dtype=float),
                repeats=True),
            Quantity(
                'mulliken',
                r'MULLIKEN POPULATION ANALYSIS \*\s*\*+([\s\S]+?)\*{10}',
                sub_parser=TextParser(quantities=population_quantities)),
            Quantity(
                'timings',
                r'\n *TIMINGS\s*\-+\s*([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=[Quantity(
                    name, rf'%s\s*\.+\s*({re_float})' % key, dtype=float, unit=ureg.s)
                    for key, name in self._timing_mapping.items()]))
        ]

        # TODO parse more properties, add to metainfo
        tddft_quantities = [
            Quantity(
                'absorption_spectrum_electric',
                r'ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS\s*'
                r'\-+[\s\S]+?\-+\n([\s\S]+?)\-{10}',
                str_operation=lambda x: np.array(
                    [v.split() for v in x.strip().split('\n')]))]

        # TODO parse more properties, add to metainfo
        mp2_quantities = [
            Quantity(
                'mp2_basis_dimension',
                r'Dimension of the basis\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'scaling_mp2_energy',
                rf'Overall scaling of the MP2 energy\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'mp2_aux_basis_dimension',
                r'Dimension of the aux\-basis\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'energy_method_current',
                rf'RI\-MP2 CORRELATION ENERGY:\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'energy_total',
                rf'MP2 TOTAL ENERGY:\s*({re_float})', dtype=float, unit=ureg.hartree)]

        def str_to_iteration_energy(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            keys = val[0]
            val = np.transpose(
                np.array([v for v in val[1:] if len(v) == len(keys)], dtype=float))
            return {keys[i]: val[i] for i in range(len(keys))}

        ci_quantities = [
            Quantity(
                'electronic_structure_method',
                r'Correlation treatment\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'single_excitations_on_off',
                r'Single excitations\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'orbital_opt_on_off',
                r'Orbital optimization\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'z_vector_calc_on_off',
                r'Calculation of Z vector\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'Brueckner_orbitals_calc_on_off',
                r'Calculation of Brueckner orbitals\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'perturbative_triple_excitations_on_off',
                r'Perturbative triple excitations\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'f12_correction_on_off',
                r'Calculation of F12 correction\s*\.+\s*(\S+)', convert=False),
            Quantity(
                'frozen_core_treatment',
                r'Frozen core treatment\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'reference_wave_function',
                r'Reference Wavefunction\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'nb_of_atomic_orbitals',
                r'Number of AO\'s\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_of_electrons',
                r'Number of electrons\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_of_correlated_electrons',
                r'Number of correlated electrons\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'integral_transformation',
                r'Integral transformation\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'level_shift_amplitude_update',
                rf'Level shift for amplitude update\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'coulomb_transformation_type',
                r'Transformation type\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'coulomb_transformation_dimension_basis',
                r'Dimension of the basis\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_internal_alpha_mol_orbitals',
                r'Number of internal alpha\-MOs\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_internal_beta_mol_orbitals',
                r'Number of internal beta\-MOs\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'pair_cutoff',
                rf'Pair cutoff\s*\.+\s*({re_float})', dtype=float),
            Quantity(
                'atomic_orbital_integral_source',
                r'AO\-integral source\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'integral_package_used',
                r'Integral package used\s*\.+\s*(.+)', flatten=False, convert=False),
            Quantity(
                'nb_alpha_pairs_included',
                r'Number of Alpha\-MO pairs included\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'nb_beta_pairs_included',
                r'Number of Beta\-MO pairs included\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'mp2_energy_spin_aa',
                rf'EMP2\(aa\)=\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'mp2_energy_spin_bb',
                rf'EMP2\(bb\)=\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'mp2_energy_spin_ab',
                rf'EMP2\(ab\)=\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'mp2_initial_guess',
                rf'E\(0\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'mp2_energy',
                rf'E\(MP2\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'mp2_total_energy',
                rf'Initial E\(tot\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'T_and_T_energy',
                rf'<T\|T>\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'total_nb_pairs_included',
                r'Number of pairs included\s*\.+\s*(\d+)', dtype=int),
            Quantity(
                'iteration_energy',
                r'(Iter\s*E\(tot\)[\s\S]+?)\-{3}',
                str_operation=str_to_iteration_energy, convert=False),
            Quantity(
                'ccsd_correlation_energy',
                rf'E\(CORR\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsd_total_energy',
                rf'E\(TOT\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'single_norm_half_ss',
                rf'Singles Norm <S\|S>\*\*1/2\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                't1_diagnostic',
                rf'T1 diagnostic\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsdt_total_triples_correction',
                rf'Triples Correction \(T\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsdt_aaa_triples_contribution',
                rf'alpha\-alpha\-alpha\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsdt_aab_triples_contribution',
                rf'alpha\-alpha\-beta\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            # typo in metainfo?
            Quantity(
                'ccsdt_aba_triples_contribution',
                rf'alpha\-beta\-beta\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsdt_bbb_triples_contribution',
                rf'beta\-beta\-beta\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsdt_final_corr_energy',
                rf'Final correlation energy\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'ccsd_final_energy',
                rf'E\(CCSD\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'energy_total',
                rf'E\(CCSD\(T\)\)\s*\.+\s*({re_float})', dtype=float, unit=ureg.hartree)]

        calculation_quantities = [
            Quantity(
                'cartesian_coordinates',
                r'CARTESIAN COORDINATES \(ANGSTROEM\)\s*\-+\s*([\s\S]+?)\n\n',
                str_operation=str_to_cartesian_coordinates),
            Quantity(
                'basis_set',
                r'\n *BASIS SET INFORMATION\s*\-+([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=basis_set_quantities)),
            Quantity(
                'auxiliary_basis_set',
                r'\n *AUXILIARY BASIS SET INFORMATION\s*\-+([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=basis_set_quantities)),
            Quantity(
                'basis_set_statistics',
                r'BASIS SET STATISTICS AND STARTUP INFO([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=basis_set_statistics_quantities)),
            Quantity(
                'self_consistent',
                r'((?:ORCA SCF|DFT GRID GENERATION)\s*\-+[\s\S]+?(?:\-{70}|\Z))',
                sub_parser=TextParser(quantities=self_consistent_quantities)),
            Quantity(
                'tddft',
                r'ORCA TD\-DFT(?:/TDA)* CALCULATION\s*\-+\s*([\s\S]+?E\(tot\).*)',
                sub_parser=TextParser(quantities=tddft_quantities)),
            Quantity(
                'mp2',
                r'ORCA MP2 CALCULATION([\s\S]+?MP2 TOTAL ENERGY:.+)',
                sub_parser=TextParser(quantities=mp2_quantities)),
            Quantity(
                'ci',
                r'ORCA\-MATRIX DRIVEN CI([\s\S]+?E\(CCSD\(T\)\).*)',
                sub_parser=TextParser(quantities=ci_quantities)
            )]

        geometry_optimization_quantities = [Quantity(
            '%s_tol' % key.lower().replace(' ', '_').replace('.', ''),
            rf'%s\s*(\w+)\s*\.+\s*({re_float})' % key, dtype=float) for key in [
                'Energy Change', 'Max. Gradient', 'RMS Gradient', 'Max. Displacement',
                'RMS Displacement']]

        geometry_optimization_quantities += [
            Quantity(
                'update_method', r'Update method\s*(\w+)\s*\.+\s*(.+)'),
            Quantity(
                'coords_choice', r'Choice of coordinates\s*(\w+)\s*\.+\s*(.+)'),
            Quantity(
                'initial_hessian', r'Initial Hessian\s*(\w+)\s*\.+\s*(.+)')]

        geometry_optimization_quantities += [
            Quantity(
                'cycle',
                r'OPTIMIZATION CYCLE\s*\d+\s*\*\s*\*+([\s\S]+?)(?:\*\s*GEOMETRY|OPTIMIZATION RUN DONE|\Z)',
                repeats=True, sub_parser=TextParser(quantities=calculation_quantities)),
            Quantity(
                'final_energy_evaluation',
                r'FINAL ENERGY EVALUATION AT THE STATIONARY POINT([\s\S]+?FINAL SINGLE POINT ENERGY.*)',
                sub_parser=TextParser(quantities=calculation_quantities))]

        self._quantities = [
            Quantity(
                'program_version',
                r'Program Version\s*([\w_.].*)', convert=False, flatten=False),
            Quantity(
                'program_svn',
                r'\(SVN:\s*\$([^$]+)\$\)\s', convert=False, flatten=False),
            Quantity(
                'program_compilation_date',
                r'\(\$Date\:\s*(\w.+?)\s*\$\)', convert=False, flatten=False),
            Quantity(
                'input_file',
                r'INPUT FILE\s*\=+([\s\S]+?)END OF INPUT',
                sub_parser=TextParser(quantities=[
                    Quantity('xc_functional', r'\d+>\s*!\s*(\S+)')])),
            Quantity(
                'single_point',
                r'\* Single Point Calculation \*\s*\*+([\s\S]+?(?:FINAL SINGLE POINT ENERGY.*|\Z))',
                sub_parser=TextParser(quantities=calculation_quantities)),
            Quantity(
                'geometry_optimization',
                r'\* Geometry Optimization Run \*\s*\*+([\s\S]+?(?:OPTIMIZATION RUN DONE|\Z))',
                sub_parser=TextParser(quantities=geometry_optimization_quantities))
        ]


class OrcaParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/orca', code_name='ORCA', code_homepage='https://orcaforum.kofo.mpg.de/',
            mainfile_contents_re=(
                r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
                r'\s+\* O   R   C   A \*\s*'
                r'\s+\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
                r'\s*'
                r'\s*--- An Ab Initio, DFT and Semiempirical electronic structure package ---\s*'))

        self._metainfo_env = m_env
        self.out_parser = OutParser()

        # TODO list adapted from old parser, is incomplete and entries may be incorrect
        self._xc_functional_map = {
            'HF': ['HS_X'],
            'HFS': ['HF_X'],
            'XALPHA': ['HF_X', 'LDA_C_XALPHA'],
            'LSD': ['LDA_X', 'LDA_C_VWN_RPA', 'VWN5_RPA'],
            'LDA': ['LDA_X', 'LDA_C_VWN_RPA', 'VWN5_RPA'],
            'VWN': ['LDA_X', 'LDA_C_VWN_RPA', 'VWN5_RPA'],
            'VWN5': ['LDA_X', 'LDA_C_VWN_RPA', 'VWN5_RPA'],
            'VWN3': ['LDA_X', 'LDA_C_VWN_3'],
            'PWLDA': ['LDA_X', 'LDA_C_PW'],
            'BNULL': ['GGA_X_B88'],
            'BVWN': ['GGA_X_B88', 'LDA_C_VWN_RPA', 'VWN5_RPA'],
            'BP': ['LDA_X', 'GGA_X_B88', 'LDA_C_VWN', 'GGA_C_P86'],
            'BP86': ['LDA_X', 'GGA_X_B88', 'LDA_C_VWN', 'GGA_C_P86'],
            'PW91': ['GGA_X_PW91', 'GGA_C_PW91'],
            'PWP': ['GGA_X_PW91', 'GGA_C_P86'],
            'MPWPW': ['GGA_X_MPW91', 'GGA_C_PW91'],
            'MPWLYP': ['GGA_X_MPW91', 'GGA_C_LYP'],
            'BLYP': ['GGA_X_B88', 'GGA_C_LYP'],
            'GP': ['GGA_X_G96', 'GGA_C_P86'],
            'GLYP': ['GGA_X_G96', 'GGA_C_LYP'],
            'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'REVPBE': ['GGA_X_PBE_R', 'GGA_C_PBE'],
            'RPBE': ['GGA_X_RPBE', 'GGA_C_PBE'],
            'OLYP': ['GGA_X_OPTX', 'GGA_C_LYP', 'GGA_XC_OPWLYP_D'],
            'OPBE': ['GGA_XC_OPBE_D'],
            'XLYP': ['GGA_XC_XLYP'],
            'B97-D': ['GGA_XC_B97_D'],
            'B97-D3': ['GGA_XC_B97_D'],
            'TPSS': ['LDA_X', 'MGGA_X_TPSS', 'LDA_C_PW', 'MGGA_C_TPSS'],
            'TPSSh': ['HYB_MGGA_XC_TPSSH'],
            'M06L': ['MGGA_X_M06_L', 'MGGA_C_M06_L'],
            'M06': ['MGGA_X_M06', 'MGGA_C_M06'],
            'M062X': ['MGGA_X_M06_2X', 'MGGA_C_M06_2X'],
            'B1LYP': ['HYB_GGA_XC_B1LYP'],
            'B3LYP': ['HYB_GGA_XC_B3LYP'],
            'B1P': ['HYB_GGA_XC_B1PW91'],
            'B3P': ['HYB_GGA_XC_B3P86'],
            'B3PW': ['HYB_GGA_XC_B3PW91'],
            'G1LYP': ['GGA_X_G96', 'GGA_C_OP_G96'],
            'PBE0': ['HYB_GGA_XC_PBEH'],
            'PW1PW': ['HYB_GGA_XC_B1PW91'],
            'PWP91_1': ['HYB_GGA_XC_B1PW91'],
            'PWP1': ['HYB_GGA_XC_B1PW91'],
            'mPW1PW': ['HYB_GGA_XC_MPW1PW'],
            'mPW1LYP': ['HYB_GGA_XC_MPWLYP1M'],
            'O3LYP': ['HYB_GGA_XC_O3LYP'],
            'X3LYP': ['HYB_GGA_XC_X3LYP'],
            'PW6B95': ['HYB_MGGA_XC_PW6B95'],
            'B97': ['HYB_GGA_XC_B97'],
            'BHANDHLYP': ['HYB_GGA_XC_BHANDHLYP'],
            'B2PLYP': ['HYB_GGA_XC_B2PLYP'],
            'MB2PLYP': ['HYB_GGA_XC_MB2PLYP'],
            'PWPB95': ['HYB_MGGA_XC_MPW1B95']}

        self._functional_type = {
            'SLATER': 'LDA', 'B88': 'GGA', 'G96': 'GGA', 'PW91': 'GGA', 'MPW': 'GGA',
            'PBE': 'GGA', 'RPBE': 'GGA', 'OPTX': 'GGA', 'X': 'GGA', 'TPSS': 'MGGA',
            'B97D': 'GGA', 'B97BECKE': 'GGA', 'SCAN': 'MGGA', 'VWN5': 'LDA', 'VWN3': 'LDA',
            'PWLDA': 'LDA', 'P86': 'GGA', 'LYP': 'GGA'}

    def parse_method(self, section):
        sec_method = self.archive.section_run[-1].m_create(Method)

        # TODO identify DFT+U
        sec_method.electronic_structure_method = 'DFT'

        scf_settings = section.get('self_consistent', {}).get('scf_settings', {})
        for key, val in scf_settings.items():
            if val is not None:
                if hasattr(val, 'units'):
                    if val.units == 'hartree':
                        val = val.to('joule').magnitude
                setattr(sec_method, 'x_orca_%s' % key, val)

        dft_grid_generation = section.get('self_consistent', {}).get('dft_grid_generation', {})
        for key, val in dft_grid_generation.items():
            if val is not None:
                setattr(sec_method, 'x_orca_%s' % key, val)

        final_grid = section.get('self_consistent', {}).get('final_grid', {})
        for key, val in final_grid.items():
            if val is not None:
                setattr(sec_method, 'x_orca_%s' % key, val)

        # xc functional
        # get type from input card
        xc_functional = self.out_parser.get('input_file', {}).get('xc_functional', '').upper()
        xc_functionals = self._xc_functional_map.get(xc_functional, [])
        # if not in map, get it from explicit exchange and correlation settings
        if len(xc_functionals) == 0:
            xc_functional_type = scf_settings.get('XC_functional_type', '')
            if xc_functional_type.startswith('Hartree-Fock'):
                xc_functionals = self._xc_functional_map.get('HF', [])

            elif xc_functional_type.startswith('DFT'):
                exchange = scf_settings.get(
                    'exchange_functional').upper().replace('-', '').split('_')[0]
                exchange_type = self._functional_type.get(exchange, None)
                if exchange_type is not None:
                    xc_functionals.append('%s_X_%s' % (exchange_type, exchange))

                if scf_settings.get('xalpha_param') is not None:
                    xc_functionals.append('LDA_X')

                correlation = scf_settings.get(
                    'correl_functional').upper().replace('-', '').split('_')[0]
                correlation_type = self._functional_type.get(correlation, None)
                if correlation_type is not None:
                    xc_functionals.append('%s_C_%s' % (correlation_type, correlation))

                lda_correlation = scf_settings.get('lda_part_of_gga_corr')
                if lda_correlation is not None:
                    xc_functionals.append('LDA_C_%s' % (lda_correlation))

                if len(xc_functionals) == 0:
                    self.logger.error('Cannot resolve xc functional', data=dict(name=xc_functional))

        for functional in xc_functionals:
            sec_xc_functionals = sec_method.m_create(XCFunctionals)
            sec_xc_functionals.XC_functional_name = functional

        for calculation_type in ['tddft', 'mp2', 'ci']:
            calculation = section.get(calculation_type)
            if calculation is None:
                continue
            method = calculation.get('electronic_structure_method')
            method = calculation_type.upper() if method is None else method
            sec_method = self.archive.section_run[-1].m_create(Method)
            sec_method.electronic_structure_method = method

            for key, val in calculation.items():
                if val is not None:
                    if hasattr(val, 'units'):
                        if val.units == 'hartree':
                            val = val.to('joule').magnitude
                    setattr(sec_method, 'x_orca_%s' % key, val)

            sec_method_ref = sec_method.m_create(MethodToMethodRefs)
            sec_method_ref.method_to_method_kind = 'starting_point'
            sec_method.method_to_method_ref = self.archive.section_run[-1].section_method[0]

        return sec_method

    def parse_system(self, section):
        sec_system = self.archive.section_run[-1].m_create(System)
        if section.get('cartesian_coordinates') is not None:
            symbols, coordinates = section.get('cartesian_coordinates')
            sec_system.atom_labels = symbols
            sec_system.atom_positions = coordinates
        sec_system.configuration_periodic_dimensions = [False] * 3
        return sec_system

    def parse_scc(self, section):
        sec_scc = self.archive.section_run[-1].m_create(SingleConfigurationCalculation)

        # TODO fix metainfo so variables take lists
        # basis sets, why are they in scc
        for kind in ['basis_set', 'auxiliary_basis_set']:
            basis_set = section.get(kind)
            if basis_set is None:
                continue
            for n in range(len(basis_set.get('basis_set', []))):
                sec_basis_set = sec_scc.m_create(BasisSet)
                for key in ['basis_set', 'basis_set_atom_labels', 'basis_set_contracted']:
                    val = basis_set.get(key)
                    if val is None:
                        continue
                    prefix = '' if key == 'basis_set_atom_labels' else kind.split('basis_set')[0]
                    metainfo_name = 'x_orca_%s%s' % (prefix, key)
                    setattr(sec_basis_set, metainfo_name, val[n])

        # gaussian basis sets
        basis_set = section.get('basis_set_statistics')
        if basis_set is not None:
            sec_basis_set = sec_scc.m_create(BasisSet)
            for key, val in basis_set.items():
                if val is None:
                    continue
                for n in range(len(val)):
                    ext = '' if n == 0 else '_aux'
                    setattr(sec_basis_set, 'x_orca_%s%s' % (key, ext), val[n])

        self_consistent = section.get('self_consistent')
        if self_consistent is None:
            return sec_scc

        scf_energy = self_consistent.get('total_scf_energy', None)
        if scf_energy is not None:
            sec_scc.energy_total = scf_energy.get('energy_total')
            energy_keys = list(self.out_parser._energy_mapping.values())
            for key, val in scf_energy.items():
                if val is not None:
                    if key in energy_keys:
                        val = val.to('joule').magnitude
                    setattr(sec_scc, 'x_orca_%s' % key, val)

        scf_iterations = self_consistent.get('scf_iterations', None)
        if scf_iterations is not None:
            for energy in scf_iterations.get('energy', []):
                sec_scf_iteration = sec_scc.m_create(ScfIteration)
                sec_scf_iteration.energy_total_scf_iteration = energy

            # why are tolerances in scf iteration
            scf_convergence = self_consistent.get('scf_convergence', {})
            for key, val in scf_convergence.items():
                if val is not None:
                    val = val.to('joule').magnitude
                    setattr(sec_scf_iteration, 'x_orca_%s' % key, val[0])
                    key = key.rstrip('_change') if 'density' in key else key
                    setattr(sec_scf_iteration, 'x_orca_%s_tolerance' % key, val[1])

        # method-specific quantities
        for calculation_type in ['tddft', 'mp2', 'ci']:
            calculation = section.get(calculation_type)
            if calculation is None:
                continue
            for key, val in calculation.items():
                if hasattr(val, 'units'):
                    if val.units == 'hartree':
                        val = val.to('joule').magnitude
                setattr(sec_scc, 'x_orca_%s' % key, val)

        # eigenvalues
        orbital_energies = self_consistent.get('orbital_energies')
        if orbital_energies is not None:
            sec_eigenvalues = sec_scc.m_create(Eigenvalues)
            orbital_energies = np.transpose(orbital_energies[:2])
            occupation = orbital_energies[1].T
            sec_eigenvalues.eigenvalues_occupation = np.reshape(
                occupation, (len(occupation), 1, len(occupation[0])))
            values = orbital_energies[2].T
            sec_eigenvalues.eigenvalues_values = np.reshape(
                values, (len(values), 1, len(values[0]))) * ureg.hartree

        # mulliken
        mulliken = self_consistent.get('mulliken')
        if mulliken is not None:
            atomic_charges = mulliken.get('atomic_charges')
            orbital_charges = mulliken.get('orbital_charges')

            for n, species in enumerate(atomic_charges.get('species')):
                sec_dos = sec_scc.m_create(Dos)
                sec_dos.x_orca_atom_species = species
                sec_dos.x_orca_mulliken_atom_charge = atomic_charges.charge[n]
                if orbital_charges is not None:
                    # TODO make orbital take list
                    sec_dos.x_orca_atom_orbital_mroc = ' '.join(
                        [c[0] for c in orbital_charges.atom[n].charge])
                    sec_dos.x_orca_mulliken_partial_orbital_charge_mroc = [
                        c[1] for c in orbital_charges.atom[n].charge]

            sec_dos.x_orca_mulliken_total_charge = atomic_charges.total_charge

        # excitations
        spectrum = section.get('tddft', {}).get('absorption_spectrum_electric')
        if spectrum is not None:
            sec_excited = sec_scc.m_create(ExcitedStates)
            spectrum = np.transpose(spectrum)
            sec_excited.x_orca_excitation_energy = (spectrum[1] / ureg.cm).to('1/m').magnitude
            sec_excited.x_orca_oscillator_strength = spectrum[3]
            sec_excited.x_orca_transition_dipole_moment_x = spectrum[5]
            sec_excited.x_orca_transition_dipole_moment_y = spectrum[6]
            sec_excited.x_orca_transition_dipole_moment_z = spectrum[7]

        # timings
        timings = self_consistent.get('timings', {})
        for key, val in timings.items():
            if val is not None:
                setattr(sec_scc, 'x_orca_%s' % key, val.magnitude)

        return sec_scc

    def parse_configurations(self):
        def parse_configuration(section):
            if section is None:
                return
            sec_method = self.parse_method(section)
            sec_system = self.parse_system(section)
            sec_scc = self.parse_scc(section)
            sec_scc.single_configuration_to_calculation_method_ref = sec_method
            sec_scc.single_configuration_calculation_to_system_ref = sec_system

        parse_configuration(self.out_parser.get('single_point'))

        geometry_optimization = self.out_parser.get('geometry_optimization')
        if geometry_optimization is not None:
            for cycle in geometry_optimization.get('cycle', []):
                parse_configuration(cycle)

            parse_configuration(geometry_optimization.get('final_energy_evaluation'))

            sec_sampling_method = self.archive.section_run[-1].m_create(SamplingMethod)
            sec_sampling_method.sampling_method = 'geometry_optimization'
            for key, val in geometry_optimization.items():
                if key in ['cycle', 'final_energy_evaluation'] or val is None:
                    continue
                if key.endswith('tol'):
                    if 'gradient' in key:
                        val[1] = (val[1] * ureg.hartree / ureg.bohr).to('joule/meter').magnitude
                    elif 'displacement' in key:
                        val[1] = (val[1] * ureg.bohr).to('meter').magnitude
                    else:
                        val[1] = (val[1] * ureg.hartree).to('joule').magnitude
                    setattr(sec_sampling_method, 'x_orca_%s_value' % key, val[1])
                    val = val[0]
                elif key in ['update_method', 'coords_choice', 'initial_hessian']:
                    setattr(sec_sampling_method, 'x_orca_%s_name' % key, ' '.join(val[1:]))
                    val = val[0]
                setattr(sec_sampling_method, 'x_orca_%s' % key, val)

    def init_parser(self, filepath, logger):
        self.out_parser.mainfile = filepath
        self.out_parser.logger = logger

    def parse(self, filepath, archive, logger):
        self.filepath = filepath
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.init_parser(filepath, logger)

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'ORCA'
        sec_run.program_basis_set_type = 'Gaussians'
        version = []
        for key in ['program_version', 'program_svn', 'program_compilation_date']:
            val = self.out_parser.get(key)
            if val is not None:
                setattr(sec_run, 'x_orca_%s' % key, val)
                version.append(val)
        sec_run.program_version = ' '.join(version)

        self.parse_configurations()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List
from abc import abstractmethod
from collections import OrderedDict
import numpy as np
from nomad.units import ureg

from nomad.datamodel.encyclopedia import (
    Material,
    Method,
)
from nomad.metainfo import Section
from nomad.normalizing.encyclopedia.basisset import get_basis_set
from nomad.normalizing.encyclopedia.context import Context
from nomad.utils import RestrictedDict
from nomad import config


class MethodNormalizer():
    """A base class that is used for processing method related information
    in the Encylopedia.
    """
    def __init__(self, entry_archive, logger):
        self.logger = logger
        self.entry_archive = entry_archive
        self.section_run = entry_archive.section_run[0]

    def method_id(self, method: Method, settings_basis_set: RestrictedDict, repr_method: Section):
        method_dict = RestrictedDict(
            mandatory_keys=[
                "program_name",
                "subsettings",
            ],
            forbidden_values=[None]
        )
        method_dict['program_name'] = self.section_run.program_name

        # The subclasses may define their own method properties that are to be
        # included here.
        subsettings = self.method_id_dict(method, settings_basis_set, repr_method)
        method_dict["subsettings"] = subsettings

        # If all required information is present, safe the hash
        try:
            method_dict.check(recursive=True)
        except (KeyError, ValueError) as e:
            self.logger.info("Could not create method hash: {}".format(e))
        else:
            method.method_id = method_dict.hash()

    @abstractmethod
    def method_id_dict(self, method: Method, settings_basis_set: RestrictedDict, repr_method: Section) -> RestrictedDict:
        pass

    def group_eos_id(self, method: Method, material: Material, repr_method: Section):
        eos_dict = RestrictedDict(
            mandatory_keys=[
                "upload_id",
                "method_id",
                "formula",
            ],
            forbidden_values=[None]
        )

        # Only calculations from the same upload are grouped
        eos_dict['upload_id'] = self.entry_archive.section_metadata.upload_id

        # Method
        eos_dict["method_id"] = method.method_id

        # The formula should be same for EoS (maybe even symmetries)
        eos_dict["formula"] = material.formula

        # Form a hash from the dictionary
        try:
            eos_dict.check(recursive=True)
        except (KeyError, ValueError) as e:
            self.logger.info("Could not create EOS hash: {}".format(e))
        else:
            method.group_eos_id = eos_dict.hash()

    def group_parametervariation_id(self, method: Method, settings_basis_set: RestrictedDict, repr_system: Section, repr_method: Section):
        # Create ordered dictionary with the values. Order is important for
        param_dict = RestrictedDict(
            mandatory_keys=[
                "upload_id",
                "program_name",
                "program_version",
                "settings_geometry",
                "subsettings",
            ],
            forbidden_values=[None]
        )

        # Only calculations from the same upload are grouped
        param_dict['upload_id'] = self.entry_archive.section_metadata.upload_id

        # The same code and functional type is required
        param_dict['program_name'] = self.section_run.program_name
        param_dict['program_version'] = self.section_run.program_version

        # Get a string representation of the geometry. It is included as the
        # geometry should remain the same during parameter variation. By simply
        # using the atom labels and positions we assume that their
        # order/translation/rotation does not change.
        geom_dict: OrderedDict = OrderedDict()
        sec_sys = repr_system
        atom_labels = sec_sys['atom_labels']
        geom_dict['atom_labels'] = ', '.join(atom_labels)
        atom_positions = sec_sys['atom_positions']
        geom_dict['atom_positions'] = np.array2string(
            atom_positions.to(ureg.angstrom).magnitude,  # convert to Angstrom
            formatter={'float_kind': lambda x: "%.6f" % x},  # type: ignore
        ).replace('\n', '')
        cell = sec_sys['lattice_vectors']
        geom_dict['simulation_cell'] = np.array2string(
            cell.to(ureg.angstrom).magnitude,  # convert to Angstrom
            formatter={'float_kind': lambda x: "%.6f" % x},  # type: ignore
        ).replace('\n', '')
        param_dict['settings_geometry'] = geom_dict

        # The subclasses may define their own method properties that are to be
        # included here.
        subsettings = self.group_parametervariation_id_dict(method, settings_basis_set, repr_method)
        param_dict["subsettings"] = subsettings

        # Form a hash from the dictionary
        try:
            param_dict.check(recursive=True)
        except (KeyError, ValueError) as e:
            self.logger.info("Could not create parameter variation hash: {}".format(e))
        else:
            method.group_parametervariation_id = param_dict.hash()

    @abstractmethod
    def group_parametervariation_id_dict(self, method: Method, settings_basis_set: RestrictedDict, repr_method: Section) -> RestrictedDict:
        pass

    def group_e_min(self) -> None:
        pass

    def group_type(self) -> None:
        pass

    @abstractmethod
    def normalize(self, context: Context) -> None:
        pass


class MethodDFTNormalizer(MethodNormalizer):
    """A base class that is used for processing method related information
    in the Encylopedia.
    """
    def core_electron_treatment(self, method: Method) -> None:
        treatment = config.services.unavailable_value
        code_name = self.section_run.program_name
        if code_name is not None:
            core_electron_treatments = {
                'VASP': 'pseudopotential',
                'FHI-aims': 'full all electron',
                'exciting': 'full all electron',
                'quantum espresso': 'pseudopotential'
            }
            treatment = core_electron_treatments.get(code_name, config.services.unavailable_value)
        method.core_electron_treatment = treatment

    def functional_long_name(self, method: Method, repr_method: Section) -> None:
        """'Long' form of exchange-correlation functional, list of components
        and parameters as a string: see
        https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional
        """
        xc_functional = MethodDFTNormalizer.functional_long_name_from_method(repr_method, self.section_run.section_method)
        if xc_functional is config.services.unavailable_value:
            self.logger.warning(
                "Metainfo for 'XC_functional' not found, and could not "
                "compose name from 'section_XC_functionals'."
            )
        method.functional_long_name = xc_functional

    @staticmethod
    def functional_long_name_from_method(repr_method: Section, methods: List[Section]):
        """'Long' form of exchange-correlation functional, list of components
        and parameters as a string: see
        https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional
        """
        linked_methods = [repr_method]
        try:
            refs = repr_method.section_method_to_method_refs
        except KeyError:
            pass
        else:
            for ref in refs:
                method_to_method_kind = ref.method_to_method_kind
                referenced_method = ref.method_to_method_ref
                if method_to_method_kind == "core_settings":
                    linked_methods.append(referenced_method)

        xc_functional = config.services.unavailable_value
        for method in linked_methods:
            try:
                section_xc_functionals = method.section_XC_functionals
            except KeyError:
                pass
            else:
                components = {}
                for component in section_xc_functionals:
                    try:
                        cname = component.XC_functional_name
                    except KeyError:
                        pass
                    else:
                        this_component = ''
                        if component.XC_functional_weight is not None:
                            this_component = str(component.XC_functional_weight) + '*'
                        this_component += cname
                        components[cname] = this_component
                result_array = []
                for name in sorted(components):
                    result_array.append(components[name])
                if len(result_array) >= 1:
                    xc_functional = '+'.join(result_array)

        return xc_functional

    def functional_type(self, method: Method) -> None:
        long_name = method.functional_long_name
        if long_name is not None:
            short_name = self.create_xc_functional_shortname(long_name)
            method.functional_type = short_name

    def method_id_dict(self, method: Method, settings_basis_set: RestrictedDict, repr_method: Section) -> RestrictedDict:
        # Extend by DFT settings.
        hash_dict = RestrictedDict(
            mandatory_keys=(
                "functional_long_name",
                "settings_basis_set",
                "scf_threshold_energy_change",
            ),
            optional_keys=(
                "smearing_kind",
                "smearing_width",
                "number_of_eigenvalues_kpoints",
            ),
            forbidden_values=[None]
        )
        # Functional settings
        hash_dict['functional_long_name'] = method.functional_long_name

        # Basis set settings
        hash_dict['settings_basis_set'] = settings_basis_set

        # k-point sampling settings if present. Add number of kpoints as
        # detected from eigenvalues. TODO: we would like to have info on the
        # _reducible_ k-point-mesh:
        #    - grid dimensions (e.g. [ 4, 4, 8 ])
        #    - or list of reducible k-points
        smearing_kind = repr_method.smearing_kind
        if smearing_kind is not None:
            hash_dict['smearing_kind'] = smearing_kind
        smearing_width = repr_method.smearing_width
        if smearing_width is not None:
            smearing_width = '%.4f' % (smearing_width)
            hash_dict['smearing_width'] = smearing_width
        try:
            scc = self.section_run.section_single_configuration_calculation[-1]
            eigenvalues = scc.section_eigenvalues
            kpt = eigenvalues[-1].eigenvalues_kpoints
        except (KeyError, IndexError):
            pass
        else:
            if kpt is not None:
                hash_dict['number_of_eigenvalues_kpoints'] = str(len(kpt))

        # SCF convergence settings
        conv_thr = repr_method.scf_threshold_energy_change
        if conv_thr is not None:
            conv_thr = '%.13f' % (conv_thr.to(ureg.rydberg).magnitude)
            hash_dict['scf_threshold_energy_change'] = conv_thr

        return hash_dict

    def group_parametervariation_id_dict(self, method: Method, settings_basis_set: RestrictedDict, repr_method: Section):
        """Dictionary containing the parameters used for convergence test
        grouping
        This is the source for generating the related hash."""
        param_dict = RestrictedDict(
            mandatory_keys=(
                "functional_long_name",
                "scf_threshold_energy_change",
            ),
            optional_keys=(
                "atoms_pseudopotentials",
            ),
            forbidden_values=[None]
        )

        # TODO: Add other DFT-specific properties
        # considered variations:
        #   - smearing kind/width
        #   - k point grids
        #   - basis set parameters
        # convergence threshold should be kept constant during convtest
        param_dict['functional_long_name'] = method.functional_long_name
        conv_thr = repr_method.scf_threshold_energy_change
        if conv_thr is not None:
            conv_thr = '%.13f' % (conv_thr.to(ureg.rydberg).magnitude)
        param_dict['scf_threshold_energy_change'] = conv_thr

        # Pseudopotentials are kept constant, if applicable
        if settings_basis_set is not None:
            pseudos = settings_basis_set.get('atoms_pseudopotentials', None)
            if pseudos is not None:
                param_dict['atoms_pseudopotentials'] = pseudos

        return param_dict

    def create_xc_functional_shortname(self, xc_longname):
        """Use lookup table to transform xc functional long- into shortname.
        """
        # Loof for "special" functional names listed in table
        """Easily editable table of 'short' XC functional names"""
        xc_functional_shortname = {
            'HF_X': 'HF',
            'HYB_GGA_XC_B3LYP5': 'hybrid-GGA',
            'HYB_GGA_XC_HSE06': 'hybrid-GGA',
            'BEEF-vdW': 'vdW-DF'
        }
        shortname = xc_functional_shortname.get(xc_longname, None)

        # If not, look into other options:
        if shortname is None:
            xc_functional_starts = {
                "LDA": "LDA",
                "GGA": "GGA",
                "HYB_GGA": "hybrid-GGA",
                "MGGA": "meta-GGA",
                "HYB_MGGA": "hybrid-meta-GGA",
                "HF": "HF"
            }
            sections = xc_longname.split("+")
            # decompose long name, this could be done more consistent with the
            # composition of the long name
            funcnames = []
            for section in sections:
                funcname = section.split('*')[-1]
                for func_start in xc_functional_starts:
                    if funcname.startswith(func_start):
                        funcnames.append(func_start)
                        break
            funcnames = set(funcnames)

            # Only one functional is defined
            # (usually for correlation and exchange)
            if len(funcnames) == 1:
                shortname = xc_functional_starts[func_start]
            # Two functionals that give a hybrid-GGA functional
            elif "GGA" in funcnames and "HF" in funcnames:
                shortname = "hybrid-GGA"

        if shortname is None:
            self.logger.info(
                "Could not find a functional shortname for xc_functional {}."
                .format(xc_longname)
            )

        return shortname

    def normalize(self, context: Context) -> None:
        # Fetch resources
        repr_method = context.representative_method
        repr_system = context.representative_system
        sec_enc = self.entry_archive.section_metadata.encyclopedia
        method = sec_enc.method
        material = sec_enc.material
        settings_basis_set = get_basis_set(context, self.entry_archive, self.logger)

        # Fill metainfo
        self.core_electron_treatment(method)
        self.functional_long_name(method, repr_method)
        self.functional_type(method)
        self.method_id(method, settings_basis_set, repr_method)
        self.group_eos_id(method, material, repr_method)
        self.group_parametervariation_id(method, settings_basis_set, repr_system, repr_method)


class MethodGWNormalizer(MethodDFTNormalizer):
    """A base class that is used for processing GW calculations.
    """
    def gw_starting_point(self, method: Method, repr_method: Section) -> None:
        try:
            ref = repr_method.section_method_to_method_refs[0]
            method_to_method_kind = ref.method_to_method_kind
            start_method = ref.method_to_method_ref
        except KeyError:
            pass
        else:
            if method_to_method_kind == "starting_point":
                methods = self.section_run.section_method
                xc_functional = MethodDFTNormalizer.functional_long_name_from_method(start_method, methods)
                method.gw_starting_point = xc_functional

    def functional_type(self, method: Method) -> None:
        method.functional_type = "GW"

    def gw_type(self, method: Method, repr_method: Section) -> None:
        method.gw_type = repr_method["electronic_structure_method"]

    def normalize(self, context: Context) -> None:
        # Fetch resources
        repr_method = context.representative_method
        sec_enc = self.entry_archive.section_metadata.encyclopedia
        method = sec_enc.method

        # Fill metainfo
        self.functional_type(method)
        self.gw_type(method, context.representative_method)
        self.gw_starting_point(method, repr_method)
